"""Get an agent's llm config argument."""

from typing import Dict, List, Tuple

from waldiez.models import WaldiezAgent, WaldiezModel

from ..models import export_agent_models


def get_agent_llm_config(
    agent: WaldiezAgent,
    agent_name: str,
    all_models: List[WaldiezModel],
    model_names: Dict[str, str],
) -> Tuple[str, str]:
    """Get the llm config argument string for one agent.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    agent_name : str
        The name of the agent.
    all_models : List[WaldiezModel]
        All the models in the flow.
    model_names : Dict[str, str]
        A mapping of model id to model name.

    Returns
    -------
    Tuple[str, str]
        The llm config argument string and extra content
        to use before the argument (the variable definition if needed).
    """
    content_before = ""
    if not agent.data.model_ids:
        # no models
        return "False", content_before
    if len(agent.data.model_ids) == 1:
        # one model
        model_id = agent.data.model_ids[0]
        model_name = model_names[model_id]
        return f"{model_name}_llm_config", content_before
    arg = f"{agent_name}_llm_config"
    content_before = "\n" + (
        export_agent_models(
            agent_model_ids=agent.data.model_ids,
            all_models=all_models,
            agent_name=agent_name,
        )
        + "\n"
    )
    return arg, content_before



"""Export the entire flow to string."""

from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

from waldiez.models import (
    Waldiez,
    WaldiezAgent,
    WaldiezChat,
    WaldiezModel,
    WaldiezSkill,
)

from ..agents import export_agent
from ..chats import export_chats, export_nested_chat
from ..models import export_models
from ..skills import export_skills
from ..utils import (
    get_comment,
    get_imports_string,
    get_logging_start_string,
    get_logging_stop_string,
    get_pylint_ignore_comment,
    get_sqlite_to_csv_call_string,
    get_sqlite_to_csv_string,
)
from .def_main import get_def_main


# pylint: disable=too-many-locals
def export_flow(
    waldiez: Waldiez,
    agents: Tuple[List[WaldiezAgent], Dict[str, str]],
    chats: Tuple[List[WaldiezChat], Dict[str, str]],
    models: Tuple[List[WaldiezModel], Dict[str, str]],
    skills: Tuple[List[WaldiezSkill], Dict[str, str]],
    output_dir: Optional[Path],
    notebook: bool,
) -> str:
    """Export the entire flow to a string.

    It contains the required imports, the model and skill definitions,
    the agent definitions, the links between agents, models and skills,
    the agents' nested chats, the chat definitions, and the actual call
    to start the chat(s).

    Parameters
    ----------
    waldiez : Waldiez
        The Waldiez instance.
    agents : Tuple[List[WaldiezAgent], Dict[str, str]]
        The agents and their names.
    chats : Tuple[List[WaldiezChat], Dict[str, str]]
        The chats and their names.
    models : Tuple[List[WaldiezModel], Dict[str, str]]
        The models and their names.
    skills : Tuple[List[WaldiezSkill], Dict[str, str]]
        The skills and their names.
    output_dir : Optional[Path]
        The output directory.
    notebook : bool
        Whether the export is for a jupyter notebook or a python script.

    Returns
    -------
    str
        The flow string.
    """
    all_agents, agent_names = agents
    all_models, model_names = models
    all_skills, skill_names = skills
    all_chats, chat_names = chats
    agent_strings = ""
    # we need to add `skipped_agent_strings` after the other agents are defined
    # for example, a group_manager needs the group members to have been defined
    skipped_agent_strings = ""
    nested_chats_strings = ""
    builtin_imports: Set[str] = {
        "import csv",
        "import os",
        "import sqlite3",
    }
    common_imports: Set[str] = {
        "from autogen import Agent",
        "from autogen import ConversableAgent",
        "from autogen import ChatResult",
        "from autogen import runtime_logging",
    }
    local_imports: Set[str] = {
        "from waldiez_api_keys import get_model_api_key",
    }
    skill_imports, _ = export_skills(
        skills=all_skills,
        skill_names=skill_names,
        output_dir=output_dir,
    )
    if len(waldiez.chats) > 1:
        common_imports.add("from autogen import initiate_chats")
    for agent in all_agents:
        agent_string, after_agent, agent_imports = export_agent(
            agent=agent,
            agent_names=agent_names,
            model_names=model_names,
            skill_names=skill_names,
            all_models=all_models,
            all_skills=all_skills,
            group_chat_members=waldiez.flow.get_group_chat_members(agent.id),
        )
        common_imports.update(agent_imports)
        if after_agent:
            skipped_agent_strings += after_agent
        if agent.agent_type == "manager":
            skipped_agent_strings += agent_string
        else:
            agent_strings += agent_string
        agent_nested_chats_string = export_nested_chat(
            agent=agent,
            agent_names=agent_names,
            all_chats=all_chats,
            chat_names=chat_names,
        )
        if agent_nested_chats_string:
            nested_chats_strings += "\n" + agent_nested_chats_string
    agent_strings += skipped_agent_strings
    models_string = export_models(
        all_models=all_models,
        model_names=model_names,
        notebook=notebook,
        output_dir=output_dir,
    )
    all_imports_string = get_imports_string(
        imports=common_imports,
        builtin_imports=builtin_imports,
        skill_imports=skill_imports,
        local_imports=local_imports,
    )
    return _combine_strings(
        waldiez=waldiez,
        imports_string=all_imports_string,
        agents_string=agent_strings,
        nested_chats_string=nested_chats_strings,
        models_string=models_string,
        agent_names=agent_names,
        chat_names=chat_names,
        notebook=notebook,
    )


# pylint: disable=too-many-arguments
def _combine_strings(
    waldiez: Waldiez,
    imports_string: str,
    agents_string: str,
    nested_chats_string: str,
    models_string: str,
    agent_names: Dict[str, str],
    chat_names: Dict[str, str],
    notebook: bool,
) -> str:
    content = get_pylint_ignore_comment(notebook)
    content += imports_string
    # content += get_comment("logging", notebook) + "\n"
    # content += get_logging_start_string(tabs=0) + "\n\n"
    content += models_string
    content += get_comment("agents", notebook) + "\n"
    content += agents_string
    if nested_chats_string:
        content += get_comment("nested", notebook) + "\n"
    content += nested_chats_string
    chats_content, additional_methods = export_chats(
        main_chats=waldiez.chats,
        agent_names=agent_names,
        chat_names=chat_names,
        tabs=0 if notebook else 1,
    )
    if additional_methods:
        while not content.endswith("\n\n"):  # pragma: no cover
            content += "\n"
        content += "\n" + additional_methods + "\n"
    content += get_sqlite_to_csv_string()
    content += get_comment("run", notebook) + "\n"
    if not notebook:
        content += get_def_main(chats_content)
    else:
        content += get_logging_start_string(tabs=0)
        content += "\n" + chats_content + "\n"
        content += get_logging_stop_string(tabs=0) + "\n"
        content += get_sqlite_to_csv_call_string(tabs=0) + "\n"
    content = content.replace("\n\n\n\n", "\n\n\n")
    return content



"""Waldiez chat related models."""

from .chat import WaldiezChat
from .chat_data import WaldiezChatData
from .chat_message import (
    WaldiezChatMessage,
    WaldiezChatMessageType,
    validate_message_dict,
)
from .chat_nested import WaldiezChatNested
from .chat_summary import WaldiezChatSummary, WaldiezChatSummaryMethod

__all__ = [
    "WaldiezChat",
    "WaldiezChatData",
    "WaldiezChatMessage",
    "WaldiezChatMessageType",
    "WaldiezChatNested",
    "WaldiezChatSummary",
    "WaldiezChatSummaryMethod",
    "validate_message_dict",
]



"""Nested chat model."""

from typing import Any, Optional

from pydantic import (
    ConfigDict,
    Field,
    ValidationInfo,
    field_validator,
    model_validator,
)
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated, Self

from ..common import WaldiezBase, WaldiezMethodName
from .chat_message import WaldiezChatMessage, validate_message_dict


class WaldiezChatNested(WaldiezBase):
    """Nested chat class.

    Attributes
    ----------
    message : WaldiezChatMessage
        The message in a nested chat (sender -> recipient).
    reply : WaldiezChatMessage
        The reply in a nested chat (recipient -> sender).
    """

    model_config = ConfigDict(
        extra="forbid",
        alias_generator=to_camel,
        populate_by_name=True,
        frozen=False,
    )

    message: Annotated[
        Optional[WaldiezChatMessage],
        Field(
            None,
            title="Message",
            description="The message in a nested chat (sender -> recipient).",
        ),
    ]
    reply: Annotated[
        Optional[WaldiezChatMessage],
        Field(
            None,
            title="Reply",
            description="The reply in a nested chat (recipient -> sender).",
        ),
    ]

    _message_content: Optional[str] = None
    _reply_content: Optional[str] = None

    @property
    def message_content(self) -> Optional[str]:
        """Get the message content."""
        return self._message_content

    @property
    def reply_content(self) -> Optional[str]:
        """Get the reply content."""
        return self._reply_content

    @field_validator("message", "reply", mode="before")
    @classmethod
    def validate_message(
        cls, value: Any, info: ValidationInfo
    ) -> WaldiezChatMessage:
        """Validate the message.

        Parameters
        ----------
        value : Any
            The value.
        info : ValidationInfo
            The validation info.

        Returns
        -------
        WaldiezChatMessage
            The validated message.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        function_name: WaldiezMethodName = (
            "nested_chat_message"
            if info.field_name == "message"
            else "nested_chat_reply"
        )
        if not value:
            return WaldiezChatMessage(
                type="none", use_carryover=False, content=None, context={}
            )
        if isinstance(value, str):
            return WaldiezChatMessage(
                type="string", use_carryover=False, content=value, context={}
            )
        if isinstance(value, dict):
            return validate_message_dict(value, function_name=function_name)
        if isinstance(value, WaldiezChatMessage):
            return validate_message_dict(
                {
                    "type": value.type,
                    "use_carryover": False,
                    "content": value.content,
                    "context": value.context,
                },
                function_name=function_name,
            )
        raise ValueError(f"Invalid message type: {type(value)}")

    @model_validator(mode="after")
    def validate_nested_chat(self) -> Self:
        """Validate the nested chat.

        Returns
        -------
        WaldiezChatNested
            The validated nested chat.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if self.message is not None:
            if self.message.type == "none":
                self._message_content = ""
            elif self.message.type == "string":
                self._message_content = self.message.content
            else:
                self._message_content = validate_message_dict(
                    value={
                        "type": "method",
                        "content": self.message.content,
                    },
                    function_name="nested_chat_message",
                    skip_definition=True,
                ).content
        if self.reply is not None:
            if self.reply.type == "none":
                self._reply_content = ""
            elif self.reply.type == "string":
                self._reply_content = self.reply.content
            else:
                self._reply_content = validate_message_dict(
                    value={
                        "type": "method",
                        "content": self.reply.content,
                    },
                    function_name="nested_chat_reply",
                    skip_definition=True,
                ).content
        return self



"""Chat related string generation functions.

Functions
---------
export_chats
    Get the chats content.
export_nested_chat
    Get the 'register_nested_chats' content.
"""

from .chats import export_chats
from .nested import export_nested_chat

__all__ = ["export_chats", "export_nested_chat"]



"""Base agent class to be inherited by all other agents."""

from .agent import WaldiezAgent, WaldiezAgentType
from .agent_data import WaldiezAgentData
from .code_execution import WaldiezAgentCodeExecutionConfig
from .linked_skill import WaldiezAgentLinkedSkill
from .nested_chat import WaldiezAgentNestedChat, WaldiezAgentNestedChatMessage
from .teachability import WaldiezAgentTeachability
from .termination_message import WaldiezAgentTerminationMessage

__all__ = [
    "WaldiezAgent",
    "WaldiezAgentCodeExecutionConfig",
    "WaldiezAgentData",
    "WaldiezAgentLinkedSkill",
    "WaldiezAgentNestedChat",
    "WaldiezAgentNestedChatMessage",
    "WaldiezAgentTeachability",
    "WaldiezAgentTerminationMessage",
    "WaldiezAgentType",
]



"""Helper functions for exporting chat data to code.

Functions
---------
export_single_chat_string
    Get the chat string when there is only one chat in the flow.
export_multiple_chats_string
    Get the chats content, when there are more than one chats in the flow.
"""

from typing import Any, Dict, List, Optional, Tuple

from waldiez.models import (
    WaldiezAgent,
    WaldiezChat,
    WaldiezChatMessage,
    WaldiezRagUser,
)

from ..utils import get_escaped_string, get_object_string


# pylint: disable=line-too-long
def export_single_chat_string(
    flow: Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent],
    agent_names: Dict[str, str],
    chat_names: Dict[str, str],
    tabs: int,
) -> Tuple[str, str]:
    """Get the chat string when there is only one chat in the flow.

    Parameters
    ----------
    flow : Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]
        The chat flow.
    agent_names : Dict[str, str]
        A mapping of agent id to agent name.
    chat_names : Dict[str, str]
        A mapping of chat id to chat name.
    tabs : int
        The number of tabs to use for indentation.

    Returns
    -------
    Tuple[str, str]
        The chat string and additional methods string if any

    Example
    -------
    ```python
    >>> from waldiez.models import WaldiezAgent, WaldiezChat, WaldiezChatData, WaldiezChatMessage
    >>> chat = WaldiezChat(
    ...     id="wc-1",
    ...     name="chat1",
    ...     description="A chat between two agents.",
    ...     tags=["chat", "chat1"],
    ...     requirements=[],
    ...     data=WaldiezChatData(
    ...         sender="wa-1",
    ...         recipient="wa-2",
    ...         message=WaldiezChatMessage(
    ...             type="string",
    ...             content="Hello, how are you?",
    ...         ),
    ...     ),
    ... )
    >>> agent_names = {"wa-1": "agent1", "wa-2": "agent2"}
    >>> chat_names = {"wc-1": "chat1"}
    >>> export_single_chat_string((chat, agent1, agent2), agent_names, chat_names, 0)
    agent1.initiate_chat(
        agent2,
        message="Hello, how are you?",
    )
    ```
    """
    tab = "    " * tabs
    chat, sender, recipient = flow
    chat_args = chat.get_chat_args(sender=sender)
    chat_args = escape_summary_args_quotes(chat_args)
    if not chat_args:
        return _get_empty_simple_chat_string(
            tab,
            chat=chat,
            sender=sender,
            recipient=recipient,
            agent_names=agent_names,
        )
    return _get_simple_chat_string(
        chat=chat,
        chat_args=chat_args,
        sender=sender,
        recipient=recipient,
        agent_names=agent_names,
        chat_names=chat_names,
        tabs=tabs,
    )


def export_multiple_chats_string(
    main_chats: List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]],
    chat_names: Dict[str, str],
    agent_names: Dict[str, str],
    tabs: int,
) -> Tuple[str, str]:
    """Get the chats content, when there are more than one chats in the flow.

    Parameters
    ----------
    main_chats : List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]
        The main chats.
    chat_names : Dict[str, str]
        A mapping of chat id to chat name.
    agent_names : Dict[str, str]
        A mapping of agent id to agent name.
    tabs : int
        The number of tabs to use for indentation.

    Returns
    -------
    Tuple[str, str]
        The main chats content and additional methods string if any.

    Example
    -------
    ```python
    >>> from waldiez.models import WaldiezAgent, WaldiezChat, WaldiezChatData, WaldiezChatMessage
    >>> chat1 = WaldiezChat(
    ...     id="wc-1",
    ...     name="chat1",
    ...     description="A chat between two agents.",
    ...     tags=["chat", "chat1"],
    ...     requirements=[],
    ...     data=WaldiezChatData(
    ...         sender="wa-1",
    ...         recipient="wa-2",
    ...         position=0,
    ...         message=WaldiezChatMessage(
    ...             type="string",
    ...             content="Hello, how are you?",
    ...         ),
    ...     ),
    ... )
    >>> chat2 = WaldiezChat(
    ...     id="wc-2",
    ...     name="chat2",
    ...     description="A chat between two agents.",
    ...     tags=["chat", "chat2"],
    ...     requirements=[],
    ...     data=WaldiezChatData(
    ...         sender="wa-2",
    ...         recipient="wa-1",
    ...         position=1,
    ...         message=WaldiezChatMessage(
    ...             type="string",
    ...             content="I am good, thank you. How about you?",
    ...         ),
    ...     ),
    ... )
    >>> agent_names = {"wa-1": "agent1", "wa-2": "agent2"}
    >>> chat_names = {"wc-1": "chat1", "wc-2": "chat2"}
    >>> export_multiple_chats_string([(chat1, agent1, agent2), (chat2, agent2, agent1)], chat_names, agent_names, 0)
    initiate_chats([
        {
            "sender": agent1,
            "recipient": agent2,
            "message": "Hello, how are you?",
        },
        {
            "sender": agent2,
            "recipient": agent1,
            "message": "I am good, thank you. How about you?",
        },
    ])
    ```
    """
    tab = "    " * tabs
    content = "\n"
    additional_methods_string = ""
    content = "initiate_chats(["
    for chat, sender, recipient in main_chats:
        chat_string, additional_methods = _get_chat_dict_string(
            chat=chat,
            chat_names=chat_names,
            sender=sender,
            recipient=recipient,
            agent_names=agent_names,
            tabs=tabs + 1,
        )
        additional_methods_string += additional_methods
        content += f"\n{tab}    {chat_string}"
    content += "\n" + "    " * tabs + "])"
    return content, additional_methods_string


def escape_summary_args_quotes(chat_args: Dict[str, Any]) -> Dict[str, Any]:
    """Escape quotes in the summary args if they are strings.

    Parameters
    ----------
    chat_args : Dict[str, Any]
        The chat arguments.

    Returns
    -------
    Dict[str, Any]
        The chat arguments with the summary prompt escaped.
    """
    if "summary_args" in chat_args and isinstance(
        chat_args["summary_args"], dict
    ):
        for key, value in chat_args["summary_args"].items():
            if isinstance(value, str):
                chat_args["summary_args"][key] = get_escaped_string(value)
    return chat_args


def _get_chat_message_string(
    chat: WaldiezChat,
    chat_names: Dict[str, str],
) -> Tuple[str, Optional[str]]:
    """Get the agent's message as a string.

    Parameters
    ----------
    chat : WaldiezChat
        The chat.
    chat_names : Dict[str, str]
        A mapping of chat id to chat name with all the chats in the flow.

    Returns
    -------
    Tuple[str, Optional[str]]
        If the message is a string, the message content and None.
        If the message is a method, the method name and the method content.
        If the message is None, 'None' and None.
    """
    if (
        not chat.message
        or chat.message.type == "none"
        or chat.message.content is None
        or chat.message_content is None
    ):
        return "None", None
    if chat.message.type == "string":
        return get_escaped_string(chat.message.content), None
    chat_name = chat_names[chat.id]
    original_function_name = "callable_message"
    method_args = "sender, recipient, context"
    function_name = f"{original_function_name}_{chat_name}"
    function_def = f"def {function_name}({method_args}):"
    return function_name, function_def + "\n" + chat.message_content + "\n"


def _get_chat_dict_string(
    chat: WaldiezChat,
    sender: WaldiezAgent,
    recipient: WaldiezAgent,
    chat_names: Dict[str, str],
    agent_names: Dict[str, str],
    tabs: int,
) -> Tuple[str, str]:
    """Get a chat dictionary string.

    If the chat message is a separate method and not a string or a lambda,
    we return the method string (definition and body) as well as the rest
    of the arguments.

    Parameters
    ----------
    chat : WaldiezChat
        The chat.
    sender : WaldiezAgent
        The sender.
    recipient : WaldiezAgent
        The recipient.
    chat_names : Dict[str, str]
        A mapping of chat id to chat name.
    agent_names : Dict[str, str]
        A mapping of agent id to agent name.
    tabs : int
        The number of tabs to use for indentation.

    Returns
    -------
    Tuple[str, str]
        The chat dictionary string and additional methods string if any.
    """
    tab = "    " * tabs
    chat_args = chat.get_chat_args(sender=sender)
    chat_args = escape_summary_args_quotes(chat_args)
    chat_string = "{"
    chat_string += "\n" + f'{tab}    "sender": {agent_names[sender.id]},'
    chat_string += "\n" + f'{tab}    "recipient": {agent_names[recipient.id]},'
    additional_methods_string = ""
    for key, value in chat_args.items():
        if isinstance(value, str):
            chat_string += "\n" + f'{tab}    "{key}": "{value}",'
        elif isinstance(value, dict):
            chat_string += (
                "\n"
                f'{tab}    "{key}": {get_object_string(value, tabs=tabs + 1)},'
            )
        else:
            chat_string += "\n" + f'{tab}    "{key}": {value},'
    if (
        sender.agent_type == "rag_user"
        and isinstance(sender, WaldiezRagUser)
        and chat.message.type == "rag_message_generator"
    ):
        message = f"{agent_names[sender.id]}.message_generator"
        chat_string += "\n" + f'{tab}    "message": {message},'
        chat_string += "\n" + tab + "},"
        return chat_string, additional_methods_string
    message, method_content = _get_chat_message_string(
        chat=chat,
        chat_names=chat_names,
    )
    if message and isinstance(chat.data.message, WaldiezChatMessage):
        message = get_escaped_string(message)
        if chat.data.message.type == "method":
            if method_content:
                additional_methods_string += "\n" + method_content
            chat_string += "\n" + f'{tab}    "message": {message},'
        elif chat.data.message.type == "string" and chat.data.message.content:
            chat_string += "\n" + f'{tab}    "message": "{message}",'
    chat_string += "\n" + tab + "},"
    return chat_string, additional_methods_string


def _get_empty_simple_chat_string(
    tab: str,
    chat: WaldiezChat,
    sender: WaldiezAgent,
    recipient: WaldiezAgent,
    agent_names: Dict[str, str],
) -> Tuple[str, str]:
    content = tab
    sender_name = agent_names[sender.id]
    recipient_name = agent_names[recipient.id]
    content += f"{sender_name}.initiate_chat(\n"
    content += tab + f"    {recipient_name},\n"
    message_arg, _ = _get_chat_message(
        tab=tab,
        chat=chat,
        chat_names={},
        sender=sender,
        sender_name=sender_name,
    )
    content += message_arg
    content += tab + ")"
    return content, ""


def _get_chat_message(
    tab: str,
    chat: WaldiezChat,
    chat_names: Dict[str, str],
    sender: WaldiezAgent,
    sender_name: str,
) -> Tuple[str, str]:
    additional_methods_string = ""
    method_content: Optional[str] = None
    if (
        sender.agent_type == "rag_user"
        and isinstance(sender, WaldiezRagUser)
        and chat.message.type == "rag_message_generator"
    ):
        message = f"{sender_name}.message_generator"
        return f"\n{tab}    message={message},", additional_methods_string
    message, method_content = _get_chat_message_string(
        chat=chat,
        chat_names=chat_names,
    )
    if message and isinstance(chat.data.message, WaldiezChatMessage):
        message = get_escaped_string(message)
        if chat.data.message.type == "method":
            additional_methods_string += (
                method_content if method_content else ""
            )
            return f"\n{tab}    message={message},", additional_methods_string
        if chat.message.type == "string" and chat.data.message.content:
            return f'\n{tab}    message="{message}",', additional_methods_string
        return "", additional_methods_string
    return "", additional_methods_string  # pragma: no cover


def _get_simple_chat_string(
    chat: WaldiezChat,
    sender: WaldiezAgent,
    recipient: WaldiezAgent,
    agent_names: Dict[str, str],
    chat_names: Dict[str, str],
    chat_args: Dict[str, Any],
    tabs: int,
) -> Tuple[str, str]:
    tab = "    " * tabs
    sender_name = agent_names[sender.id]
    recipient_name = agent_names[recipient.id]
    chat_string = f"{sender_name}.initiate_chat(\n"
    chat_string += f"{tab}    {recipient_name},"
    for key, value in chat_args.items():
        if isinstance(value, str):
            chat_string += f'\n{tab}    {key}="{value}",'
        elif isinstance(value, dict):
            chat_string += (
                f"\n{tab}    {key}={get_object_string(value, tabs + 1)},"
            )
        else:
            chat_string += f"\n{tab}    {key}={value},"
    message_arg, additional_methods_string = _get_chat_message(
        tab=tab,
        chat=chat,
        chat_names=chat_names,
        sender=sender,
        sender_name=sender_name,
    )
    chat_string += message_arg
    chat_string += f"\n{tab})"
    return chat_string, additional_methods_string



"""Waldiez flow model."""

import uuid
from typing import List, Optional, Tuple

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from ..agents import WaldiezAgent
from ..chat import WaldiezChat
from ..common import WaldiezBase, now
from .flow_data import WaldiezFlowData


class WaldiezFlow(WaldiezBase):
    """Flow data class.

    Attributes
    ----------
    id : str
        The ID of the flow.
    type : Literal["flow"], optional
        The type of the "node" in a graph: "flow".
    name : str
        The name of the flow.
    description : str
        The description of the flow.
    tags : List[str]
        The tags of the flow.
    requirements : List[str]
        The requirements of the flow.
    storage_id : str
        The storage ID of the flow (ignored, UI related).
    created_at : str
        The date and time when the flow was created.
    updated_at : str
        The date and time when the flow was last updated.
    data : WaldiezFlowData
        The data of the flow. See `WaldiezFlowData`.
    """

    id: Annotated[
        str,
        Field(
            description="The ID of the flow",
            title="ID",
            default_factory=uuid.uuid4,
        ),
    ]
    type: Annotated[
        Literal["flow"],
        Field(
            "flow",
            description="The type of the 'node' in a graph",
            title="Type",
        ),
    ]
    name: Annotated[
        str,
        Field(
            ...,
            description="The name of the flow",
            title="Name",
        ),
    ]
    description: Annotated[
        str,
        Field(
            ...,
            description="The description of the flow",
            title="Description",
        ),
    ]
    tags: Annotated[
        List[str],
        Field(
            description="The tags of the flow",
            title="Tags",
            default_factory=list,
        ),
    ]
    requirements: Annotated[
        List[str],
        Field(
            description="The requirements of the flow",
            title="Requirements",
            default_factory=list,
        ),
    ]
    data: Annotated[
        WaldiezFlowData,
        Field(
            ...,
            description="The data of the flow",
            title="Data",
        ),
    ]
    storage_id: Annotated[
        str,
        Field(
            uuid.uuid4(),
            description="The storage ID of the flow (ignored, UI related)",
            title="Storage ID",
            alias="storageId",
        ),
    ]
    created_at: Annotated[
        str,
        Field(
            default_factory=now,
            title="Created At",
            description="The date and time when the flow was created.",
        ),
    ]
    updated_at: Annotated[
        str,
        Field(
            default_factory=now,
            title="Updated At",
            description="The date and time when the flow was last updated.",
        ),
    ]
    _ordered_flow: Optional[
        List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]
    ] = None

    @property
    def ordered_flow(
        self,
    ) -> List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]:
        """Get the ordered flow."""
        if not self._ordered_flow:
            self._ordered_flow = self._get_flow_order()
        return self._ordered_flow

    def get_agent_by_id(self, agent_id: str) -> WaldiezAgent:
        """Get the agent by ID.

        Parameters
        ----------
        agent_id : str
            The ID of the agent.

        Returns
        -------
        WaldiezAgent
            The agent.

        Raises
        ------
        ValueError
            If the agent with the given ID is not found.
        """
        for user in self.data.agents.users:
            if user.id == agent_id:
                return user
        for assistant in self.data.agents.assistants:
            if assistant.id == agent_id:
                return assistant
        for manager in self.data.agents.managers:
            if manager.id == agent_id:
                return manager
        for rag_user in self.data.agents.rag_users:
            if rag_user.id == agent_id:
                return rag_user
        raise ValueError(f"Agent with ID {agent_id} not found.")

    def _get_flow_order(
        self,
    ) -> List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]:
        """Get the ordered flow."""
        # in the chats, there is the 'order' field, we use this,
        # we only keep the ones with order >=0
        # and sort them by this property
        ordered_flow: List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]] = []
        sorted_chats_by_order = sorted(
            self.data.chats, key=lambda chat: chat.data.order
        )
        for chat in sorted_chats_by_order:
            if chat.data.order < 0:
                continue
            source = self.get_agent_by_id(chat.source)
            target = self.get_agent_by_id(chat.target)
            ordered_flow.append((chat, source, target))
        if not ordered_flow:
            if len(self.data.chats) == 1:
                chat = self.data.chats[0]
                source = self.get_agent_by_id(chat.source)
                target = self.get_agent_by_id(chat.target)
                ordered_flow.append((chat, source, target))
        return ordered_flow

    def get_agent_connections(
        self, agent_id: str, all_chats: bool = True
    ) -> List[str]:
        """Get the agent connections.

        Parameters
        ----------
        agent_id : str
            The ID of the agent.
        all_chats : bool, optional
            If True, get the connections from all the chats, otherwise
            get the connections from the ordered flow (main chat flow).

        Returns
        -------
        List[str]
            The list of agent ids that the agent with the given ID connects to.
        """
        connections = []
        if all_chats:
            for chat in self.data.chats:
                if chat.source == agent_id:
                    connections.append(chat.target)
                if chat.target == agent_id:
                    connections.append(chat.source)
        else:
            for _, source, target in self.ordered_flow:
                if source.id == agent_id:
                    connections.append(target.id)
                if target.id == agent_id:
                    connections.append(source.id)
        return connections

    def get_group_chat_members(
        self, group_manager_id: str
    ) -> List[WaldiezAgent]:
        """Get the group chat members.

        Parameters
        ----------
        group_manager_id : str
            The ID of the group manager.

        Returns
        -------
        List[WaldiezAgent]
            The list of group chat
        """
        agent = self.get_agent_by_id(group_manager_id)
        if agent.agent_type != "manager":
            return []
        connections = self.get_agent_connections(group_manager_id)
        return [self.get_agent_by_id(member_id) for member_id in connections]

    def _validate_agent_connections(self) -> None:
        for agent in self.data.agents.members:
            if not any(
                agent.id in (chat.source, chat.target)
                for chat in self.data.chats
            ):
                raise ValueError(
                    f"Agent {agent.id} ({agent.name}) "
                    "does not connect to any other node."
                )
            # already covered above
            # if agent.agent_type == "manager":
            #     chat_member_ids = self.get_agent_connections(agent.id)
            #     if not chat_member_ids:
            #         raise ValueError(
            #             f"Manager's {agent.id} group chat has no members."
            #         )

    @model_validator(mode="after")
    def validate_flow(self) -> Self:
        """Flow validation.

        - unique node ids
        - there are at least two agents
        - all the agents connect to at least one other agent
        - all the linked agent skills are found in the flow
        - all the linked agent models are found in the flow
        - all the managers have at least one member in the chat group
        - the ordered flow (chats with position >=0) is not empty
        - all agents' code execution config functions exist in the flow skills

        Returns
        -------
        WaldiezFlow
            The validated flow.

        Raises
        ------
        ValueError
            If the ordered flow is empty.
            If the model IDs are not unique.
            If the skill IDs are not unique.
            If the agents do not connect to any other node.
            If the manager's group chat has no members.
        """
        if not self.ordered_flow:
            raise ValueError("The ordered flow is empty.")
        model_ids = [model.id for model in self.data.models]
        if len(model_ids) != len(set(model_ids)):
            raise ValueError("Model IDs must be unique.")
        skills_ids = [skill.id for skill in self.data.skills]
        if len(skills_ids) != len(set(skills_ids)):
            raise ValueError("Skill IDs must be unique.")
        self.data.agents.validate_flow(model_ids, skills_ids)
        self._validate_agent_connections()
        return self



"""Export the chats content."""

from typing import Dict, List, Tuple

from waldiez.models import WaldiezAgent, WaldiezChat

from .helpers import export_multiple_chats_string, export_single_chat_string


def export_chats(
    main_chats: List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]],
    agent_names: Dict[str, str],
    chat_names: Dict[str, str],
    tabs: int,
) -> Tuple[str, str]:
    """Get the chats content.

    Parameters
    ----------
    main_chats : List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]
        The main flow chats.
    agent_names : Dict[str, str]
        A mapping of agent id to agent name.
    chat_names : Dict[str, str]
        A mapping of chat id to chat name.
    tabs : int
        The number of tabs to use for indentation.

    Returns
    -------
    Tuple[str, str]
        The chats content and additional methods string if any.
    """
    if len(main_chats) == 1:
        return export_single_chat_string(
            flow=main_chats[0],
            agent_names=agent_names,
            chat_names=chat_names,
            tabs=tabs,
        )
    return export_multiple_chats_string(
        main_chats=main_chats,
        chat_names=chat_names,
        agent_names=agent_names,
        tabs=tabs,
    )



"""Get the main function (if exporting to python)."""

from ..utils import (
    get_logging_start_string,
    get_logging_stop_string,
    get_sqlite_to_csv_call_string,
)


def get_def_main(waldiez_chats: str) -> str:
    """Get the main function.

    When exporting to python, waldiez_chats string will be the
    content of the main function. It contains either a
    `{sender.initiate_chat(recipient, ...)}` (if there is only one chat)
    or `initiate_chats([..])`, with the list of chats to initiate.

    Parameters
    ----------
    waldiez_chats : str
        The content of the main function.

    Returns
    -------
    str
        The main function.
    """
    content = """def main():
    # type: () -> Union[ChatResult, List[ChatResult]]
    \"\"\"Start chatting.\"\"\"
"""
    content += get_logging_start_string(1)
    content += f"    results = {waldiez_chats}" + "\n"
    content += get_logging_stop_string(1) + "\n"
    content += get_sqlite_to_csv_call_string(1) + "\n"
    content += "    return results\n"
    return content



"""Path check utility functions."""

import os
from pathlib import Path
from typing import Optional

# pylint: disable=broad-except


def _check_local_path(string: str) -> Optional[Path]:
    """Check if a string is a local path.

    Parameters
    ----------
    string : str
        The string to check.

    Returns
    -------
    bool
        True if the path is a local path.
    """
    try:
        path = Path(string).resolve()
    except Exception:  # pragma: no cover
        return None
    if path.exists():
        return path
    return None


def get_path_string(string: str) -> str:
    """Get the path string.

    Parameters
    ----------
    string : str
        The string to check.

    Returns
    -------
    str
        The local path string.
    """
    # On windows, we get paths like "C:\path\to\file"
    # if so, let's try to avoid invalid escape sequences
    if not _check_local_path(string):
        return string
    if os.name == "nt":  # pragma: no cover
        return f"r'{string}'"
    return f"{Path(string).resolve()}"



"""Waldiez Model Data."""

from typing import Dict, Optional

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..common import WaldiezBase

WaldiezModelAPIType = Literal[
    "openai",
    "azure",
    "google",
    "anthropic",
    "mistral",
    "groq",
    "together",
    "nim",
    "other",
]


class WaldiezModelPrice(WaldiezBase):
    """Model Price.

    Attributes
    ----------
    prompt_price_per_1k : float
        The prompt price per 1k tokens.
    completion_token_price_per_1k : float
        The completion token price per 1k tokens.
    """

    prompt_price_per_1k: Annotated[
        Optional[float], Field(None, alias="promptPricePer1k")
    ]
    completion_token_price_per_1k: Annotated[
        Optional[float], Field(None, alias="completionTokenPricePer1k")
    ]


# pylint: disable=line-too-long
class WaldiezModelData(WaldiezBase):
    """Waldiez Model Data.

    Attributes
    ----------
    base_url : Optional[str]
        The base url of the model, by default None.
    api_key : Optional[str]
        The api key to use with the model, by default None.
    api_type : Literal["openai","azure","google","anthropic","mistral","groq","together","nim","other"]
        The api type of the model.
    api_version : Optional[str]
        The api version of the model, by default None.
    temperature : Optional[float]
        The temperature of the model, by default None.
    top_p : Optional[float]
        The top p of the model, by default None.
    max_tokens : Optional[int]
        The max tokens of the model, by default None.
    default_headers : Dict[str, str]
        The default headers of the model.
    price : Optional[WaldiezModelPrice]
        The price of the model, by default None.
    """

    base_url: Annotated[
        Optional[str],
        Field(
            None,
            title="Base URL",
            description="The base url of the model",
            alias="baseUrl",
        ),
    ]
    api_key: Annotated[Optional[str], Field(None, alias="apiKey")]
    api_type: Annotated[WaldiezModelAPIType, Field("other", alias="apiType")]
    api_version: Annotated[Optional[str], Field(None, alias="apiVersion")]
    temperature: Annotated[Optional[float], Field(None, alias="temperature")]
    top_p: Annotated[Optional[float], Field(None, alias="topP")]
    max_tokens: Annotated[Optional[int], Field(None, alias="maxTokens")]
    default_headers: Annotated[
        Dict[str, str], Field(alias="defaultHeaders", default_factory=dict)
    ]
    price: Annotated[Optional[WaldiezModelPrice], Field(None)]



"""Agent strings generation.."""

from typing import Dict, List, Set, Tuple

from waldiez.models import WaldiezAgent, WaldiezModel, WaldiezSkill

from ..utils import get_escaped_string
from .agent_skills import get_agent_skill_registrations
from .code_execution import get_agent_code_execution_config
from .group_manager import get_group_manager_extras
from .llm_config import get_agent_llm_config
from .rag_user import get_rag_user_extras
from .termination_message import get_is_termination_message


def get_agent_class_name(agent: WaldiezAgent) -> str:
    """Get the agent class name.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.

    Returns
    -------
    str
        The agent class name.
    """
    if agent.data.is_multimodal:
        return "MultimodalConversableAgent"
    if agent.agent_type == "assistant":
        return "AssistantAgent"
    if agent.agent_type == "user":
        return "UserProxyAgent"
    if agent.agent_type == "manager":
        return "GroupChatManager"
    if agent.agent_type == "rag_user":
        return "RetrieveUserProxyAgent"
    return "ConversableAgent"  # pragma: no cover


def get_agent_imports(agent_class: str) -> Set[str]:
    """Get the imports needed for the agent.

    Parameters
    ----------
    agent_class : str
        The agent class name.

    Returns
    -------
    Set[str]
        The imports needed for the agent.
    """
    imports = set()
    if agent_class == "AssistantAgent":
        imports.add("from autogen import AssistantAgent")
    elif agent_class == "UserProxyAgent":
        imports.add("from autogen import UserProxyAgent")
    elif agent_class == "GroupChatManager":
        imports.add("from autogen import GroupChatManager")
        imports.add("from autogen import GroupChat")
    elif agent_class == "RetrieveUserProxyAgent":
        imports.add(
            "from autogen.agentchat.contrib.retrieve_user_proxy_agent "
            "import RetrieveUserProxyAgent"
        )
    elif agent_class == "MultimodalConversableAgent":
        imports.add(
            "from autogen.agentchat.contrib.multimodal_conversable_agent "
            "import MultimodalConversableAgent"
        )
    return imports


def get_system_message_arg(agent: WaldiezAgent) -> str:
    """Get the system message argument.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.

    Returns
    -------
    str
        The system message argument.
    """
    if not agent.data.system_message:
        return ""
    return (
        "\n    "
        f'system_message="{get_escaped_string(agent.data.system_message)}",'
    )


# pylint: disable=too-many-locals, unused-argument
def export_agent(
    agent: WaldiezAgent,
    agent_names: Dict[str, str],
    model_names: Dict[str, str],
    skill_names: Dict[str, str],
    all_models: List[WaldiezModel],
    all_skills: List[WaldiezSkill],
    group_chat_members: List[WaldiezAgent],
) -> Tuple[str, str, Set[str]]:
    """Export the agent to a string.

    If the agent's `is_termination_msg` is a method,
    the function definition and content will be included in the string.
    So it could be like:
    ```python
    ...
    def is_termination_message_{agent_name}(message):
        return ....
    ...
    agent_name = AssistantAgent(
        ...
        is_termination_msg=is_termination_message_{agent_name},
        ...
    )
    ```
    The same goes for any additional `before the agent` contents in the cases
    of a group chat manager (define the `GroupChat` first),
    a retrieve user proxy agent (define the retrieve/db config first),
    or the agent's `code_execution`: define the
    `DockerCommandLineCodeExecutor`/`LocalCommandLineCodeExecutor` first.

    The agent's `skill_registrations` and/or nested chats if any,
    should be added to the final string after all the agents are defined.
    definition. Example:
    ```python
        ...
        agent_name = AssistantAgent(
            ...
        )
        ...
        register_function(
            {skill_name},
            caller={caller_name},
            executor={executor_name},
        )
        ...
    ```

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    agent_names : Dict[str, str]
        A mapping of agent id to agent name.
    model_names : Dict[str, str]
        A mapping of model id to model name.
    skill_names : Dict[str, str]
        A mapping of skill id to skill name.
    all_models : List[WaldiezModel]
        All the models in the flow.
    all_skills : List[WaldiezSkill]
        All the skills in the flow.
    group_chat_members : List[WaldiezAgent]
        The group chat members.

    Returns
    -------
    Tuple[str, str, Set[str], Set[str]]
        A tuple containing:
        - The string representation of the agent (w additional content before),
        - Extra content to be added after the agents are defined.
        - Needed imports (autogen.x, db/RAG or code execution related) if any.
    """
    imports: Set[str] = set()
    agent_name = agent_names[agent.id]
    before_agent_string = ""
    after_agent_string = ""
    before_manager, group_chat_arg = get_group_manager_extras(
        agent, group_chat_members, agent_names
    )
    if before_manager:
        before_agent_string += before_manager
    before_rag, retrieve_arg, rag_imports = get_rag_user_extras(
        agent, agent_name, model_names
    )
    if before_rag:
        before_agent_string += before_rag
    imports.update(rag_imports)
    is_termination_message, termination_function = get_is_termination_message(
        agent, agent_name
    )
    if termination_function:
        before_agent_string += termination_function
    executor, config_arg, coding_import = get_agent_code_execution_config(
        agent=agent, agent_name=agent_name, skill_names=skill_names
    )
    if executor:
        before_agent_string += executor
    if coding_import:
        imports.add(f"from autogen.coding import {coding_import}")
    agent_class = get_agent_class_name(agent)
    imports.update(get_agent_imports(agent_class))
    if agent.data.skills:
        imports.add("from autogen import register_function")
    default_auto_reply: str = "None"
    if agent.data.agent_default_auto_reply:
        default_auto_reply = (
            f'"{get_escaped_string(agent.data.agent_default_auto_reply)}"'
        )
    agent_llm_config_arg, llm_config_string = get_agent_llm_config(
        agent=agent,
        agent_name=agent_name,
        all_models=all_models,
        model_names=model_names,
    )
    before_agent_string += llm_config_string
    agent_str = f"""{agent_name} = {agent_class}(
    name="{agent_name}",
    description="{agent.description}",
    llm_config={agent_llm_config_arg},{(get_system_message_arg(agent))}
    human_input_mode="{agent.data.human_input_mode}",
    max_consecutive_auto_reply={agent.data.max_consecutive_auto_reply},
    default_auto_reply={default_auto_reply},
    code_execution_config={config_arg},
    is_termination_msg={is_termination_message},{group_chat_arg}{retrieve_arg}
)
"""
    agent_skill_registrations = get_agent_skill_registrations(
        agent, agent_names, all_skills, skill_names
    )
    if before_agent_string:
        agent_str = before_agent_string + "\n" + agent_str
    if agent_skill_registrations:
        after_agent_string = "\n" + agent_skill_registrations + "\n"
    return (
        agent_str,
        after_agent_string,
        imports,
    )



"""Group chat manger agent."""

from .group_manager import WaldiezGroupManager
from .group_manager_data import WaldiezGroupManagerData
from .speakers import (
    WaldiezGroupManagerSpeakers,
    WaldiezGroupManagerSpeakersSelectionMethod,
    WaldiezGroupManagerSpeakersSelectionMode,
    WaldiezGroupManagerSpeakersTransitionsType,
)

__all__ = [
    "WaldiezGroupManager",
    "WaldiezGroupManagerData",
    "WaldiezGroupManagerSpeakers",
    "WaldiezGroupManagerSpeakersSelectionMethod",
    "WaldiezGroupManagerSpeakersSelectionMode",
    "WaldiezGroupManagerSpeakersTransitionsType",
]



"""Waldiez Flow runner.

Run a waldiez flow.
The flow is first converted to an autogen flow with agents, chats and skills.
We then chown to temporary directory, call the flow's `main()` and
return the results. Before running the flow, any additional environment
variables specified in the waldiez file are set.
"""

import datetime
import importlib.util
import io
import os
import shutil
import subprocess  # nosemgrep # nosec
import sys
import tempfile
from contextlib import contextmanager
from contextvars import ContextVar
from pathlib import Path
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Type,
    Union,
)

from .exporter import WaldiezExporter
from .models.waldiez import Waldiez

if TYPE_CHECKING:
    from autogen import ChatResult  # type: ignore

    from .io import WaldiezIOStream


@contextmanager
def _chdir(to: Union[str, Path]) -> Iterator[None]:
    """Change the current working directory in a context.

    Parameters
    ----------
    to : Union[str, Path]
        The directory to change to.

    Yields
    ------
    Iterator[None]
        The context manager.
    """
    old_cwd = str(os.getcwd())
    os.chdir(to)
    try:
        yield
    finally:
        os.chdir(old_cwd)


class WaldiezRunner:
    """Waldiez runner class."""

    def __init__(
        self, waldiez: Waldiez, file_path: Optional[Union[str, Path]] = None
    ) -> None:
        """Initialize the Waldiez manager."""
        self._waldiez = waldiez
        self._running = False
        self._file_path = file_path
        self._stream: ContextVar[Optional["WaldiezIOStream"]] = ContextVar(
            "waldiez_stream", default=None
        )
        self._exporter = WaldiezExporter(waldiez)

    @classmethod
    def load(
        cls,
        waldiez_file: Union[str, Path],
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        requirements: Optional[List[str]] = None,
    ) -> "WaldiezRunner":
        """Create a WaldiezRunner instance from a file.

        Parameters
        ----------
        waldiez_file : Union[str, Path]
            The file path.
        name : Optional[str], optional
            The name of the Waldiez, by default None.
        description : Optional[str], optional
            The description of the Waldiez, by default None.
        tags : Optional[List[str]], optional
            The tags of the Waldiez, by default None.
        requirements : Optional[List[str]], optional
            The requirements of the Waldiez, by default None.

        Returns
        -------
        WaldiezRunner
            The Waldiez runner instance.

        Raises
        ------
        FileNotFoundError
            If the file is not found.
        RuntimeError
            If the file is not a valid Waldiez file.
        """
        waldiez = Waldiez.load(
            waldiez_file,
            name=name,
            description=description,
            tags=tags,
            requirements=requirements,
        )
        return cls(waldiez, file_path=waldiez_file)

    def __enter__(
        self,
    ) -> "WaldiezRunner":
        """Enter the context manager."""
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        """Exit the context manager."""
        if self._running:
            self._running = False
        token = self._stream.get()
        if token is not None:
            self._stream.reset(token)
            del token

    @property
    def waldiez(self) -> Waldiez:
        """Get the Waldiez instance."""
        return self._waldiez

    @property
    def running(self) -> bool:
        """Get the running status."""
        return self._running

    def _get_print_function(self) -> Callable[..., None]:
        """Get the print function."""
        token = self._stream.get()
        if token is not None:
            return token.print
        return print

    def _install_requirements(self) -> None:
        """Install the requirements for the flow."""
        extra_requirements = set(
            req for req in self.waldiez.requirements if req not in sys.modules
        )
        if extra_requirements:
            print_function = self._get_print_function()
            # pylint: disable=inconsistent-quotes
            print_function(
                f"Installing requirements: {', '.join(extra_requirements)}"
            )
            pip_install = [sys.executable, "-m", "pip", "install"]
            if not in_virtualenv():
                pip_install.append("--user")
            pip_install.extend(extra_requirements)
            with subprocess.Popen(
                pip_install,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            ) as proc:
                if proc.stdout:
                    for line in io.TextIOWrapper(proc.stdout, encoding="utf-8"):
                        print_function(line.strip())
                if proc.stderr:
                    for line in io.TextIOWrapper(proc.stderr, encoding="utf-8"):
                        print_function(line.strip())
            print_function(
                "Requirements installed.\n"
                "NOTE: If new packages were added and you are using Jupyter, "
                "you might need to restart the kernel."
            )

    def _after_run(
        self, temp_dir: Path, output_path: Optional[Union[str, Path]]
    ) -> None:
        if output_path:
            destination_dir = Path(output_path).parent
            destination_dir = (
                destination_dir
                / "waldiez_out"
                / datetime.datetime.now().strftime("%Y%m%d%H%M%S")
            )
            destination_dir.mkdir(parents=True, exist_ok=True)
            # copy the contents of the temp dir to the destination dir
            self._get_print_function()(
                f"Copying the results to {destination_dir}"
            )
            for item in temp_dir.iterdir():
                # skip cache files
                if (
                    item.name.startswith("__pycache__")
                    or item.name.endswith(".pyc")
                    or item == ".cache"
                ):
                    continue
                if item.is_file():
                    shutil.copy(item, destination_dir)
                else:
                    shutil.copytree(item, destination_dir / item.name)
        shutil.rmtree(temp_dir)

    def _set_env_vars(self) -> Dict[str, str]:
        """Set environment variables and return the old ones (if any)."""
        old_vars: Dict[str, str] = {}
        for var_key, var_value in self.waldiez.get_flow_env_vars():
            if var_key:
                current = os.environ.get(var_key, "")
                old_vars[var_key] = current
                os.environ[var_key] = var_value
        return old_vars

    @staticmethod
    def _reset_env_vars(old_vars: Dict[str, str]) -> None:
        """Reset the environment variables."""
        for var_key, var_value in old_vars.items():
            if not var_value:
                os.environ.pop(var_key, "")
            else:
                os.environ[var_key] = var_value

    def _do_run(
        self,
        output_path: Optional[Union[str, Path]],
        uploads_root: Optional[Union[str, Path]],
    ) -> Union["ChatResult", List["ChatResult"]]:
        """Run the Waldiez workflow.

        Parameters
        ----------
        output_path : Optional[Union[str, Path]]
            The output path.
        uploads_root : Optional[Union[str, Path]]
            The runtime uploads root.

        Returns
        -------
        Union[ChatResult, List[ChatResult]]
            The result(s) of the chat(s).
        """
        results: Union["ChatResult", List["ChatResult"]] = []
        if not uploads_root:
            uploads_root = Path(tempfile.mkdtemp())
        else:
            uploads_root = Path(uploads_root)
        if not uploads_root.exists():
            uploads_root.mkdir(parents=True)
        temp_dir = Path(tempfile.mkdtemp())
        file_name = "flow.py" if not output_path else Path(output_path).name
        if file_name.endswith((".json", ".waldiez")):
            file_name = file_name.replace(".json", ".py").replace(
                ".waldiez", ".py"
            )
        if not file_name.endswith(".py"):
            file_name += ".py"
        module_name = file_name.replace(".py", "")
        with _chdir(to=temp_dir):
            self._exporter.export(Path(file_name))
            spec = importlib.util.spec_from_file_location(
                module_name, temp_dir / file_name
            )
            if not spec or not spec.loader:
                raise ImportError("Could not import the flow")
            sys.path.insert(0, str(temp_dir))
            old_vars = self._set_env_vars()
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            print_function = self._get_print_function()
            print_function("Starting workflow...")
            results = module.main()
            sys.path.pop(0)
            self._reset_env_vars(old_vars)
        self._after_run(temp_dir, output_path)
        return results

    def _run(
        self,
        output_path: Optional[Union[str, Path]],
        uploads_root: Optional[Union[str, Path]],
    ) -> Union["ChatResult", List["ChatResult"]]:
        self._install_requirements()
        token = self._stream.get()
        if token is not None:
            # pylint: disable=import-outside-toplevel
            from .io import WaldiezIOStream

            with WaldiezIOStream.set_default(token):
                return self._do_run(output_path, uploads_root)
        return self._do_run(output_path, uploads_root)

    def run(
        self,
        stream: Optional["WaldiezIOStream"] = None,
        output_path: Optional[Union[str, Path]] = None,
        uploads_root: Optional[Union[str, Path]] = None,
    ) -> Union["ChatResult", List["ChatResult"]]:
        """Run the Waldiez workflow.

        Parameters
        ----------
        stream : Optional[WaldiezIOStream], optional
            The stream to use, by default None.
        output_path : Optional[Union[str, Path]], optional
            The output path, by default None.
        uploads_root : Optional[Union[str, Path]], optional
            The uploads root, to get user-uploaded files, by default None.

        Returns
        -------
        Union[ChatResult, List[ChatResult]]
            The result(s) of the chat(s).

        Raises
        ------
        RuntimeError
            If the workflow is already running.
        """
        if self._running is True:
            raise RuntimeError("Workflow already running")
        self._running = True
        token = self._stream.set(stream)
        file_path = output_path or self._file_path
        try:
            return self._run(file_path, uploads_root)
        finally:
            self._running = False
            self._stream.reset(token)
            del token


def in_virtualenv() -> bool:
    """Check if we are inside a virtualenv.

    Returns
    -------
    bool
        True if inside a virtualenv, False otherwise.
    """
    return hasattr(sys, "real_prefix") or (
        hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
    )



"""Check for conflicts with 'autogen-agentchat' package."""

# pylint: disable=line-too-long

import sys
from importlib.metadata import PackageNotFoundError, version


# fmt: off
def check_conflicts() -> None:  # pragma: no cover
    """Check for conflicts with 'autogen-agentchat' package."""
    try:
        version("autogen-agentchat")
        print(
            "Conflict detected: 'autogen-agentchat' is installed, which conflicts with 'ag2'.\n"
            "Please uninstall 'autogen-agentchat': pip uninstall -y autogen-agentchat \n"
            "And install 'ag2' (or 'waldiez') again: pip install --force ag2"
        )
        sys.exit(1)
    except PackageNotFoundError:
        pass

# fmt: on



"""Assistant agent data model."""

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..agent import WaldiezAgentData


class WaldiezAssistantData(WaldiezAgentData):
    """Assistant agent data class.

    The data for an agent with `human_input_mode` set to `"ALWAYS"` as default.
    See the parent's docs (`WaldiezAgentData`) for the rest of the properties.

    Attributes
    ----------
    human_input_mode : Literal["ALWAYS", "NEVER", "TERMINATE"]
        The human input mode, Defaults to `NEVER`
    """

    human_input_mode: Annotated[
        Literal["ALWAYS", "NEVER", "TERMINATE"],
        Field(
            "NEVER",
            title="Human input mode",
            description="The human input mode, Defaults to `NEVER`",
            alias="humanInputMode",
        ),
    ]



"""Skills/tools related string generation functions.

Functions
---------
get_agent_skill_registration
    Get an agent's skill registration string.
export_skills
    Get the skills content and secrets.
"""

from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple, Union

from waldiez.models import WaldiezSkill

from ..utils import get_escaped_string


def get_agent_skill_registration(
    caller_name: str,
    executor_name: str,
    skill_name: str,
    skill_description: str,
) -> str:
    """Get the agent skill string and secrets.

    Parameters
    ----------
    caller_name : str
        The name of the caller (agent).
    executor_name : str
        The name of the executor (agent).
    skill_name : str
        The name of the skill.
    skill_description : str
        The skill description.

    Returns
    -------
    str
        The agent skill string.

    Example
    -------
    ```python
    >>> get_agent_skill_registration(
    ...     caller_name="agent1",
    ...     executor_name="agent2",
    ...     skill_name="skill1",
    ...     skill_description="A skill that does something.",
    ... )
    register_function(
        skill1,
        caller=agent1,
        executor=agent2,
        name="skill1",
        description="A skill that does something.",
    )
    ```
    """
    skill_description = get_escaped_string(skill_description)
    content = f"""register_function(
    {skill_name},
    caller={caller_name},
    executor={executor_name},
    name="{skill_name}",
    description="{skill_description}",
)"""
    return content


def _write_skill_secrets(
    skill: WaldiezSkill,
    skill_name: str,
    output_dir: Path,
) -> None:
    """Write the skill secrets to a file.

    Parameters
    ----------
    skill : WaldiezSkill
        The skill.
    skill_name : str
        The name of the skill.
    output_dir : Path
        The output directory to save the secrets to.
    """
    if not skill.secrets:
        return
    secrets_file = output_dir / f"{skill_name}_secrets.py"
    with secrets_file.open("w", encoding="utf-8") as f:
        f.write('"""Secrets for the skill."""\n')
        f.write("from os import environ\n\n")
        for key, value in skill.secrets.items():
            f.write(f'environ["{key}"] = "{value}"\n')


def export_skills(
    skills: List[WaldiezSkill],
    skill_names: Dict[str, str],
    output_dir: Optional[Union[str, Path]] = None,
) -> Tuple[Set[str], Set[Tuple[str, str]]]:
    """Get the skills' contents and secrets.

    If `output_dir` is provided, the contents are saved to that directory.

    Parameters
    ----------
    skills : List[WaldiezSkill]
        The skills.
    skill_names : Dict[str, str]
        The skill names.
    output_dir : Optional[Union[str, Path]]
        The output directory to save the skills to.

    Returns
    -------
    Tuple[Set[str], Set[Tuple[str, str]]]
        - The skill imports to use in the main file.
        - The skill secrets to set as environment variables.

    Example
    -------
    ```python
    >>> from waldiez.models import WaldiezSkill, WaldiezSkillData
    >>> skill1 = WaldiezSkill(
    ...     id="ws-1",
    ...     name="skill1",
    ...     description="A skill that does something.",
    ...     tags=["skill", "skill1"],
    ...     requirements=[],
    ...     data=WaldiezSkillData(
    ...         content="def skill1():\\n    pass",
    ...         secrets={"API_KEY": "1234567890"},
    ... )
    >>> skill_names = {"ws-1": "skill1"}
    >>> export_skills([skill1], skill_names, None)
    ({'from skill1 import skill1'}, {('API_KEY', '1234567890')})
    ```
    """
    skill_imports: Set[str] = set()
    skill_secrets: Set[Tuple[str, str]] = set()
    for skill in skills:
        skill_name = skill_names[skill.id]
        skill_secrets.update(skill.secrets.items())
        if not output_dir:
            skill_imports.add(f"from {skill_name} import {skill_name}")
            continue
        if not isinstance(output_dir, Path):
            output_dir = Path(output_dir)
        if not skill.secrets:
            skill_imports.add(f"from {skill_name} import {skill_name}")
        else:
            # have the secrets before the skill
            skill_imports.add(
                f"import {skill_name}_secrets  # noqa\n"
                f"from {skill_name} import {skill_name}"
            )
        _write_skill_secrets(skill, skill_name, output_dir)
        skill_file = output_dir / f"{skill_name}.py"
        with skill_file.open("w", encoding="utf-8") as f:
            f.write(skill.content)
    return skill_imports, skill_secrets



"""Model/LLM related string generation functions.

Functions
---------
export_models
    Get the string representations of the LLM configs.
"""

from pathlib import Path
from typing import Dict, List, Optional

from waldiez.models import WaldiezModel

from ..utils import get_comment, get_object_string


def export_models(
    all_models: List[WaldiezModel],
    model_names: Dict[str, str],
    notebook: bool,
    output_dir: Optional[Path] = None,
) -> str:
    """Get the string representations of the LLM configs.

    Parameters
    ----------
    all_models : List[WaldiezModel]
        The models.
    model_names : Dict[str, str]
        A mapping of model ids to model names.
    notebook : bool
        Whether to export the string for a jupyter notebook.
    output_dir : Optional[Path]
        The output directory to write the api keys.
    Returns
    -------
    str
        The string representation of the models.

    Example
    -------
    ```python
    >>> from waldiez.models import WaldiezModel, WaldiezModelData
    >>> api_key = "1234567890"
    ... model = WaldiezModel(
    ...     id="wm-1",
    ...     name="llama3.1"  ,
    ...     description="A model for llamas :P.",
    ...     tags=["llama", "llama3.1"],
    ...     requirements=[],
    ...     data=WaldiezModelData(
    ...         base_url="https://example.com/v1",
    ...         api_key=api_key,
    ...         api_type="openai",
    ...         temperature=0.5,
    ...         price={
    ...             "prompt_price_per_1k": 0.0001,
    ...             "completion_token_price_per_1k": 0.0002,
    ...         },
    ...     ),
    ... )
    >>> model_names = {"wm-1": "llama3_1"}
    >>> export_models([model], model_names, True)

    # # Models
    llama3_1_llm_config = {
        "config_list": [{
            "model": "llama3_1",
            "base_url": "https://example.com/v1",
            "api_key": get_model_api_key("llama3_1"),
            "api_type": "openai",
            "temperature": 0.5,
            "price": [0.0001, 0.0002],
        }]
    }
    ```
    """
    content = get_comment("models", notebook) + "\n"
    for model in all_models:
        model_name = model_names[model.id]
        llm_config = model.get_llm_config()
        llm_config["api_key"] = f'get_model_api_key("{model_name}")'
        model_dict_str = get_object_string(llm_config, tabs=2)
        model_dict_str = model_dict_str.replace(
            f'"get_model_api_key("{model_name}")"',
            f'get_model_api_key("{model_name}")',
        )
        content += f"{model_name}_llm_config = " + "{\n"
        content += '    "config_list": [\n'
        content += f"        {model_dict_str}\n"
        content += "    ]\n"
        content += "}\n"
    if output_dir:
        write_api_keys(all_models, model_names, output_dir)
    return content


def export_agent_models(
    agent_model_ids: List[str],
    all_models: List[WaldiezModel],
    agent_name: str,
) -> str:
    """Get the string representations of the agent's registered models.

    Parameters
    ----------
    agent_model_ids : List[str]
        The model ids registered to the agent.
    all_models : List[WaldiezModel]
        All the models in the flow.
    agent_name : str
        The name of the agent.

    Returns
    -------
    str
        The agent's llm config string.

    Example
    -------
    ```python
    >>> from waldiez.models import WaldiezModel, WaldiezModelData
    >>> model1 = WaldiezModel(
    ...     id="wm-1",
    ...     name="llama3.1"  ,
    ...     description="A model for llamas :P.",
    ...     tags=["llama", "llama3.1"],
    ...     requirements=[],
    ...     data=WaldiezModelData(
    ...         base_url="https://example.com/v1",
    ...         api_key="1234567890",
    ...         api_type="openai",
    ...         temperature=0.5,
    ...         price={
    ...             "prompt_price_per_1k": 0.0001,
    ...             "completion_token_price_per_1k": 0.0002,
    ...         },
    ...     ),
    ... )
    >>> model2 = WaldiezModel(
    ...     id="wm-2",
    ...     name="llama3.2"  ,
    ...     description="A model for llamas :P.",
    ...     tags=["llama", "llama3.2"],
    ...     requirements=[],
    ...     data=WaldiezModelData(
    ...         base_url="https://example.com/v1",
    ...        api_key="1234567890",
    ...         api_type="openai",
    ...         temperature=0.5,
    ...         price={
    ...             "prompt_price_per_1k": 0.0001,
    ...             "completion_token_price_per_1k": 0.0002,
    ...         },
    ...     ),
    ... )
    >>> all_models = [model1, model2]
    >>> agent_model_ids = ["wm-1", "wm-2"]
    >>> export_agent_models(agent_model_ids, all_models, "llama_agent")

    llama_agent_llm_config = {
        "config_list": [
            {
                "model": "llama3.1",
                "base_url": "https://example.com/v1",
                "api_key": "1234567890",
                "api_type": "openai",
                "temperature": 0.5,
                "price": [0.0001, 0.0002],
            },
            {
                "model": "llama3.2",
                "base_url": "https://example.com/v1",
                "api_key": "1234567890",
                "api_type": "openai",
                "temperature": 0.5,
                "price": [0.0001, 0.0002],
            },
        ]
    }
    ```
    """
    content = f"{agent_name}_llm_config = " + "{\n"
    content += '    "config_list": [\n'
    for model_id in agent_model_ids:
        model = next((m for m in all_models if m.id == model_id), None)
        if model is not None:
            llm_config = model.get_llm_config()
            model_dict_str = get_object_string(llm_config, tabs=2)
            content += f"        {model_dict_str},\n"
    content += "    ]\n"
    content += "}\n"
    return content


def write_api_keys(
    all_models: List[WaldiezModel],
    model_names: Dict[str, str],
    output_dir: Path,
) -> None:
    """Write the api keys to a separate file.

    Parameters
    ----------
    all_models : List[WaldiezModel]
        All the models in the flow.
    model_names : Dict[str, str]
        A mapping of model ids to model names.
    output_dir : Path
        The output directory to write the api keys.
    """
    # example call: llama3_1_api_key = get_model_api_key("llama3_1")
    api_keys_content = '''
"""API keys for the models."""

__ALL_MODEL_API_KEYS__ = {
'''
    for model in all_models:
        model_name = model_names[model.id]
        api_keys_content += f'    "{model_name}": "{model.api_key}",\n'
    api_keys_content += "}\n"

    api_keys_content += '''

def get_model_api_key(model_name: str) -> str:
    """Get the api key for the model.

    Parameters
    ----------
    model_name : str
        The name of the model.

    Returns
    -------
    str
        The api key for the model.
    """
    return __ALL_MODEL_API_KEYS__.get(model_name, "")
'''

    with open(output_dir / "waldiez_api_keys.py", "w", encoding="utf-8") as f:
        f.write(api_keys_content)



"""Generic utils to be used for exporting."""

from .comments import comment, get_comment, get_pylint_ignore_comment
from .importing import add_autogen_dot_import, get_imports_string
from .logging_utils import (
    get_logging_start_string,
    get_logging_stop_string,
    get_sqlite_to_csv_call_string,
    get_sqlite_to_csv_string,
)
from .method_utils import get_method_string
from .naming import (
    get_escaped_string,
    get_valid_instance_name,
    get_valid_python_variable_name,
)
from .object_string import get_object_string
from .path_check import get_path_string

__all__ = [
    "add_autogen_dot_import",
    "comment",
    "get_logging_start_string",
    "get_logging_stop_string",
    "get_path_string",
    "get_pylint_ignore_comment",
    "get_sqlite_to_csv_string",
    "get_sqlite_to_csv_call_string",
    "get_imports_string",
    "get_comment",
    "get_escaped_string",
    "get_method_string",
    "get_object_string",
    "get_valid_instance_name",
    "get_valid_python_variable_name",
]



"""Waldiez Agent Skill Model."""

from pydantic import Field
from typing_extensions import Annotated

from ...common import WaldiezBase


class WaldiezAgentLinkedSkill(WaldiezBase):
    """Waldiez Agent Linked Skill.

    Attributes
    ----------
    id : str
        The id of the skill to use.
    executor_id: str
        The id of the agent to use that skill.
    """

    id: Annotated[
        str, Field(..., title="ID", description="The id of the skill to use.")
    ]
    executor_id: Annotated[
        str,
        Field(
            ...,
            title="Executor ID",
            description="The id of the agent to use that skill.",
        ),
    ]



"""Chat data model."""

from typing import Any, Dict, Optional, Union

from pydantic import ConfigDict, Field, field_validator, model_validator
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated, Self

from ..common import WaldiezBase
from .chat_message import WaldiezChatMessage, validate_message_dict
from .chat_nested import WaldiezChatNested
from .chat_summary import WaldiezChatSummary


class WaldiezChatData(WaldiezBase):
    """Chat data class.

    Attributes
    ----------
    name : str
        The name of the chat.
    source : str
        The source of the chat (sender).
    target : str
        The target of the chat (recipient).
    description : str
        The description of the chat.
    position : int
        The position of the chat. Ignored (UI related).
    order : int
        The of the chat. If negative, ignored.
    clear_history : Optional[bool], optional
        Whether to clear the chat history, by default None.
    message : Union[str, WaldiezChatMessage]
        The message of the chat.
    nested_chat : WaldiezChatNested
        The nested chat config.
    summary : WaldiezChatSummary
        The summary method and options for the chat.
    max_turns : Optional[int]
        The maximum number of turns for the chat, by default None (no limit).
    silent : Optional[bool], optional
        Whether to run the chat silently, by default None (ignored).
    summary_args : Optional[Dict[str, Any]]
        The summary args to use in autogen.
    real_source : Optional[str]
        The real source of the chat (overrides the source).
    real_target : Optional[str]
        The real target of the chat (overrides the target).

    Functions
    ---------
    validate_message(value: Any)
        Validate the message.
    validate_summary_method(value: Optional[WaldiezChatSummaryMethod])
        Validate the summary method.
    serialize_summary_method(value: Any, info: FieldSerializationInfo)
        Serialize summary method.
    get_chat_args()
        Get the chat arguments to use in autogen.
    """

    model_config = ConfigDict(
        extra="forbid",
        alias_generator=to_camel,
        populate_by_name=True,
        frozen=False,
    )

    name: Annotated[
        str, Field(..., title="Name", description="The name of the chat.")
    ]
    source: Annotated[
        str,
        Field(
            ...,
            title="Source",
            description="The source of the chat (sender).",
        ),
    ]
    target: Annotated[
        str,
        Field(
            ...,
            title="Target",
            description="The target of the chat (recipient).",
        ),
    ]
    description: Annotated[
        str,
        Field(
            ...,
            title="Description",
            description="The description of the chat.",
        ),
    ]
    position: Annotated[
        int,
        Field(
            -1,
            title="Position",
            description="The position of the chat in the flow (Ignored).",
        ),
    ]
    order: Annotated[
        int,
        Field(
            -1,
            title="Order",
            description="The order of the chat in the flow.",
        ),
    ]
    clear_history: Annotated[
        Optional[bool],
        Field(
            None,
            alias="clearHistory",
            title="Clear History",
            description="Whether to clear the chat history.",
        ),
    ]
    message: Annotated[
        Union[str, WaldiezChatMessage],
        Field(
            title="Message",
            description="The message of the chat.",
            default_factory=WaldiezChatMessage,
        ),
    ]
    nested_chat: Annotated[
        WaldiezChatNested,
        Field(
            title="Nested Chat",
            description="The nested chat.",
            alias="nestedChat",
            default_factory=WaldiezChatNested,
        ),
    ]
    summary: Annotated[
        WaldiezChatSummary,
        Field(
            default_factory=WaldiezChatSummary,
            title="Summary",
            description="The summary method options for the chat.",
        ),
    ]
    max_turns: Annotated[
        Optional[int],
        Field(
            None,
            alias="maxTurns",
            title="Max Turns",
            description="The maximum number of turns for the chat.",
        ),
    ]
    silent: Annotated[
        Optional[bool],
        Field(
            None,
            title="Silent",
            description="Whether to run the chat silently.",
        ),
    ]
    real_source: Annotated[
        Optional[str],
        Field(
            None,
            alias="realSource",
            title="Real Source",
            description="The real source of the chat (overrides the source).",
        ),
    ]
    real_target: Annotated[
        Optional[str],
        Field(
            None,
            alias="realTarget",
            title="Real Target",
            description="The real target of the chat (overrides the target).",
        ),
    ]

    _message_content: Optional[str] = None

    @property
    def message_content(self) -> Optional[str]:
        """Get the message content."""
        return self._message_content

    @model_validator(mode="after")
    def validate_chat_data(self) -> Self:
        """Validate the chat data.

        Returns
        -------
        WaldiezChatData
            The validated chat data.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if isinstance(self.message, WaldiezChatMessage):
            if self.message.type == "none":
                self._message_content = None
            elif self.message.type == "string":
                self._message_content = self.message.content
            else:
                self._message_content = validate_message_dict(
                    value={
                        "type": self.message.type,
                        "content": self.message.content,
                        "use_carryover": self.message.use_carryover,
                    },
                    function_name="callable_message",
                    skip_definition=True,
                ).content
        return self

    @field_validator("message", mode="before")
    @classmethod
    def validate_message(cls, value: Any) -> WaldiezChatMessage:
        """Validate the message.

        Parameters
        ----------
        value : Any
            The message value.

        Returns
        -------
        WaldiezChatMessage
            The validated message value.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if value is None:
            return WaldiezChatMessage(
                type="none", use_carryover=False, content=None, context={}
            )
        if isinstance(value, str):
            return WaldiezChatMessage(
                type="string", use_carryover=False, content=value, context={}
            )
        if isinstance(value, dict):
            return validate_message_dict(
                value, function_name="callable_message"
            )
        if isinstance(value, WaldiezChatMessage):
            return validate_message_dict(
                value={
                    "type": value.type,
                    "use_carryover": value.use_carryover,
                    "content": value.content,
                    "context": value.context,
                },
                function_name="callable_message",
            )
        return WaldiezChatMessage(
            type="none", use_carryover=False, content=None, context={}
        )

    @property
    def summary_args(self) -> Optional[Dict[str, Any]]:
        """Get the summary args."""
        if self.summary.method not in (
            "reflection_with_llm",
            "reflectionWithLlm",
        ):
            return None
        args: Dict[str, Any] = {}
        if self.summary.prompt:
            args["summary_prompt"] = self.summary.prompt
        if self.summary.args:
            args.update(self.summary.args)
        return args

    def _get_context_args(self) -> Dict[str, Any]:
        """Get the context arguments to use in autogen.

        Returns
        -------
        Dict[str, Any]
            The dictionary to use for generating the kwargs.
        """
        extra_args: Dict[str, Any] = {}
        if isinstance(self.message, WaldiezChatMessage):
            for key, value in self.message.context.items():
                if str(value).lower() in ("none", "null"):
                    extra_args[key] = None
                elif str(value).isdigit():
                    extra_args[key] = int(value)
                elif str(value).replace(".", "").isdigit():
                    try:
                        extra_args[key] = float(value)
                    except ValueError:  # pragma: no cover
                        extra_args[key] = value
                else:
                    extra_args[key] = value
        return extra_args

    def get_chat_args(self) -> Dict[str, Any]:
        """Get the chat arguments to use in autogen.

        Returns
        -------
        Dict[str, Any]
            The dictionary to pass as kwargs.
        """
        args: Dict[str, Any] = {}
        if self.summary.method:
            args["summary_method"] = self.summary.method
        if self.summary_args:
            args["summary_args"] = self.summary_args
        if isinstance(self.max_turns, int) and self.max_turns > 0:
            args["max_turns"] = self.max_turns
        if isinstance(self.clear_history, bool):
            args["clear_history"] = self.clear_history
        if isinstance(self.silent, bool):
            args["silent"] = self.silent
        args.update(self._get_context_args())
        return args



"""Importing related string generation functions.

Functions
---------
add_autogen_dot_import
    Add an autogen dot import (from autogen.{x.y} import {z}).
get_imports_string
    Get the imports for the whole file/flow.
"""

from typing import Dict, List, Optional, Set, Tuple

DEFAULT_TYPING_IMPORTS = {
    "Any",
    "Callable",
    "Dict",
    "List",
    "Optional",
    "Tuple",
    "Union",
}


def add_autogen_dot_import(
    current_imports: Dict[str, List[str]], new_import: Tuple[str, str]
) -> Dict[str, List[str]]:
    """Add an autogen dot import (from autogen.{x} import {y}).

    Parameters
    ----------
    current_imports : Dict[str, List[str]]
        The current autogen dot imports.
    new_import : Tuple[str, str]
        The new import.

    Returns
    -------
    Dict[str, List[str]]
        The updated imports.

    Example
    -------
    ```python
    >>> current_imports = {"a": ["b", "c"], "d": ["e"]}
    >>> new_import = ("a", "f")
    >>> add_autogen_dot_import(current_imports, new_import)
    {'a': ['b', 'c', 'f'], 'd': ['e']}
    # and the final string would be:
    from autogen.a import b, c, f
    from autogen.d import e
    ```
    """
    dot_part, module_part = new_import
    if not module_part:
        return current_imports
    imports_copy = current_imports.copy()
    if dot_part not in current_imports:
        imports_copy[dot_part] = []
    imports_copy[dot_part].append(module_part)
    return imports_copy


def get_imports_string(
    imports: Set[str],
    skill_imports: Set[str],
    typing_imports: Optional[Set[str]] = None,
    builtin_imports: Optional[Set[str]] = None,
    local_imports: Optional[Set[str]] = None,
) -> str:
    """Get the imports.

    Parameters
    ----------
    imports : Set[str]
        The flow imports (e.g. from autogen.{x[y.z]} import {w}).
    skill_imports : Set[str]
        The skill imports.
    typing_imports : Set[str], optional
        The typing imports, by default None.
    builtin_imports : Set[str], optional
        The builtin imports, by default None.
    local_imports : Set[str], optional
        The local imports (like for getting model api keys), by default None.

    Returns
    -------
    str
        The imports string.

    Example
    -------
    ```python
    >>> autogen_imports = {"from autogen import a", "from autogen.b import c"}
    >>> skill_imports = {"from skill_name import skill_name"}
    >>> get_imports_string(autogen_imports, skill_imports)

    from typing import Any, Callable, Dict, List, Optional, Tuple, Union
    from typing_extensions import Annotated

    from autogen import a
    from autogen.b import c

    from skill_name import skill_name'
    ```
    """
    if not typing_imports:
        typing_imports = DEFAULT_TYPING_IMPORTS
    if not builtin_imports:
        builtin_imports = set()
    if not local_imports:
        local_imports = set()
    string = _get_builtin_imports_string(builtin_imports, typing_imports)
    string += _get_third_party_imports_string(imports)
    if local_imports:
        string += "\n\n" + "\n".join(sorted(local_imports)) + "\n"
    string += _get_skill_imports_string(skill_imports)
    string = "\n\n".join([line for line in string.split("\n\n") if line])
    while not string.endswith("\n\n"):
        string += "\n"
    return string


# pylint: disable=line-too-long
def _get_builtin_imports_string(
    builtin_imports: Set[str],
    typing_imports: Set[str],
    include_annotations: bool = True,
) -> str:
    """Get the builtin imports."""
    imports = []
    from_imports = []
    for imp in sorted(builtin_imports):
        if imp.startswith("from "):
            from_imports.append(imp)
        elif imp.startswith("import "):
            imports.append(imp)
        else:
            imports.append(f"import {imp}")
    if typing_imports:
        without_from_typing = []
        for typing_import in typing_imports:
            if typing_import.startswith("from typing import "):
                without_from_typing.extend(
                    typing_import.split("from typing import ")[1].split(", ")
                )
            else:
                without_from_typing.append(typing_import)
        from_imports.append(
            "from typing import "
            + ", ".join(sorted(without_from_typing))
            + "  # noqa"
        )
        if include_annotations:
            from_imports.append("from typing_extensions import Annotated")
    string = (
        "\n".join(sorted(imports)) + "\n\n" + "\n".join(sorted(from_imports))
    )
    return string


def _get_autogen_import(import_string: str) -> List[str]:
    """Get the autogen import.

    In case the import is a "full" import statement,
    we keep only the module to import.
    """
    things = []
    if import_string.startswith("from autogen import "):
        import_part = import_string.split("from autogen import ")[1]
        things = import_part.split(", ")
    return things


def _prepare_imports(
    autogen_imports: Set[str],
    autogen_dot_imports: Dict[str, List[str]],
    other_imports: Set[str],
) -> Tuple[List[str], Dict[str, List[str]]]:
    plain_imports = []  # plain `import {x}`
    autogen_imports_list = []  # from autogen import {y}
    for autogen_import in autogen_imports:
        autogen_imports_list.extend(_get_autogen_import(autogen_import))
    from_imports_dict: Dict[str, List[str]] = {
        "autogen": autogen_imports_list,
    }
    # from autogen.{z} import {w}  # z could be "a.b.c.."
    for autogen_dot_package, autogen_dot_modules in autogen_dot_imports.items():
        sub_package = f"autogen.{autogen_dot_package}"
        if sub_package not in from_imports_dict:
            from_imports_dict[sub_package] = []
        from_imports_dict[sub_package].extend(autogen_dot_modules)
    for imp in other_imports:
        if imp.startswith("from "):
            line_parts = imp.split("from ")
            package = line_parts[1].split(" import ")[0]
            if package not in from_imports_dict:
                from_imports_dict[package] = []
            import_part = line_parts[1].split(" import ")[1]
            things = import_part.split(", ")
            from_imports_dict[package].extend(things)
        elif imp.startswith("import "):
            plain_imports.append(imp)
        else:
            plain_imports.append(f"import {imp}")
    plain_imports.sort()
    return plain_imports, from_imports_dict


def _get_autogen_imports(
    imports: Set[str],
) -> Tuple[Set[str], Dict[str, List[str]], Set[str]]:
    """Get the autogen imports."""
    autogen_imports = set()
    _autogen_dot_imports: Dict[str, List[str]] = {}
    remaining_imports: Set[str] = set()
    for imp in imports:
        if imp.startswith("from autogen import "):
            autogen_imports.add(imp)
        elif imp.startswith("from autogen."):
            parts = imp.split("from autogen.")[1].split(" import ")
            if len(parts) == 2:
                package = parts[0]
                module = parts[1]
                if package not in _autogen_dot_imports:
                    _autogen_dot_imports[package] = []
                _autogen_dot_imports[package].append(module)
        else:
            remaining_imports.add(imp)
    # sort the autogen imports
    autogen_imports = set(sorted(list(autogen_imports)))
    autogen_dot_imports = {}
    # sort the autogen dot imports (both keys and values)
    sorted_keys = sorted(_autogen_dot_imports.keys())
    for key in sorted_keys:
        autogen_dot_imports[key] = sorted(_autogen_dot_imports[key])
    return autogen_imports, autogen_dot_imports, remaining_imports


def _get_third_party_imports_string(imports: Set[str]) -> str:
    """Get the third party imports."""
    autogen_imports, autogen_dot_imports, rest = _get_autogen_imports(imports)
    plain_imports, from_imports_dict = _prepare_imports(
        autogen_imports=autogen_imports,
        autogen_dot_imports=autogen_dot_imports,
        other_imports=rest,
    )
    from_imports = []
    # pylint: disable=inconsistent-quotes
    for package, modules in from_imports_dict.items():
        # remove duplicates
        modules = sorted(set(modules))
        if modules:
            from_imports.append(f"from {package} import {', '.join(modules)}")
    string = (
        "\n\n" + "\n".join(plain_imports) + "\n\n" + "\n".join(from_imports)
    )
    return string


def _get_skill_imports_string(skill_imports: Set[str]) -> str:
    """Get the skill imports."""
    if not skill_imports:
        return ""
    string = "\n\n" + "\n".join(sorted(skill_imports)) + "\n"
    return string



"""Agent models."""

from .agent import (
    WaldiezAgent,
    WaldiezAgentCodeExecutionConfig,
    WaldiezAgentData,
    WaldiezAgentLinkedSkill,
    WaldiezAgentNestedChat,
    WaldiezAgentNestedChatMessage,
    WaldiezAgentTeachability,
    WaldiezAgentTerminationMessage,
    WaldiezAgentType,
)
from .agents import WaldiezAgents
from .assistant import WaldiezAssistant, WaldiezAssistantData
from .group_manager import (
    WaldiezGroupManager,
    WaldiezGroupManagerData,
    WaldiezGroupManagerSpeakers,
    WaldiezGroupManagerSpeakersSelectionMethod,
    WaldiezGroupManagerSpeakersSelectionMode,
    WaldiezGroupManagerSpeakersTransitionsType,
)
from .rag_user import (
    WaldiezRagUser,
    WaldiezRagUserChunkMode,
    WaldiezRagUserData,
    WaldiezRagUserModels,
    WaldiezRagUserRetrieveConfig,
    WaldiezRagUserTask,
    WaldiezRagUserVectorDb,
    WaldiezRagUserVectorDbConfig,
)
from .user_proxy import WaldiezUserProxy, WaldiezUserProxyData

__all__ = [
    "WaldiezAgent",
    "WaldiezAgentType",
    "WaldiezAgents",
    "WaldiezAssistant",
    "WaldiezAssistantData",
    "WaldiezAgentCodeExecutionConfig",
    "WaldiezAgentData",
    "WaldiezAgentLinkedSkill",
    "WaldiezAgentNestedChat",
    "WaldiezAgentNestedChatMessage",
    "WaldiezAgentTeachability",
    "WaldiezAgentTerminationMessage",
    "WaldiezGroupManager",
    "WaldiezGroupManagerData",
    "WaldiezGroupManagerSpeakers",
    "WaldiezGroupManagerSpeakersSelectionMethod",
    "WaldiezGroupManagerSpeakersSelectionMode",
    "WaldiezGroupManagerSpeakersTransitionsType",
    "WaldiezRagUser",
    "WaldiezRagUserData",
    "WaldiezRagUserModels",
    "WaldiezUserProxy",
    "WaldiezUserProxyData",
    "WaldiezRagUserRetrieveConfig",
    "WaldiezRagUserTask",
    "WaldiezRagUserChunkMode",
    "WaldiezRagUserVectorDb",
    "WaldiezRagUserVectorDbConfig",
]



"""Comment related string generation functions.
Functions
---------
comment
    Get a comment string.
get_comment
    Get a comment string for some common keys (notebook headings).
get_pylint_ignore_comment
    Get the pylint ignore comment string.
"""

from typing import List, Optional

from typing_extensions import Literal

PYLINT_RULES = [
    "line-too-long",
    "unknown-option-value",
    "unused-argument",
    "unused-import",
    "invalid-name",
    "import-error",
    "inconsistent-quotes",
    "missing-function-docstring",
    "missing-param-doc",
    "missing-return-doc",
]


def comment(notebook: bool, hashtags: int = 1) -> str:
    """Get the comment string.

    Parameters
    ----------
    notebook : bool
        Whether the comment is for a notebook or not.
    hashtags : int, optional
        The number of hashtags (for notebooks), by default 1.

    Returns
    -------
    str
        The comment string.
    Example
    -------
    ```python
    >>> comment(True, 2)
    '## '
    >>> comment(False)
    '# '
    ```
    """
    content = "# "
    if notebook:
        content += "#" * hashtags + " "
    return content


def get_comment(
    key: Literal["agents", "skills", "models", "nested", "run", "logging"],
    for_notebook: bool,
) -> str:
    """Get a comment string for some common keys.

    The key is a heading (in a notebook) or just a comment (in a script).

    Parameters
    ----------
    key : Literal["agents", "skills", "models", "nested", "run", "logging"]
        The key.
    for_notebook : bool
        Whether the comment is for a notebook.

    Returns
    -------
    str
        The comment string.

    Example
    -------
    ```python
    >>> get_comment("agents", True)

    '## Agents'
    >>> get_comment("skills", False)

    '# Skills'
    ```
    """
    # pylint: disable=too-many-return-statements
    if key == "agents":
        return "\n" + comment(for_notebook, 2) + "Agents\n"
    if key == "skills":
        return "\n" + comment(for_notebook, 2) + "Skills\n"
    if key == "models":
        return "\n" + comment(for_notebook, 2) + "Models\n"
    if key == "nested":
        return "\n" + comment(for_notebook, 2) + "Nested Chats\n"
    if key == "run":
        return "\n" + comment(for_notebook, 2) + "Run the flow\n"
    if key == "logging":
        return "\n" + comment(for_notebook, 2) + "Start Logging\n"
    return comment(for_notebook)


def get_pylint_ignore_comment(
    notebook: bool, rules: Optional[List[str]] = None
) -> str:
    """Get the pylint ignore comment string.

    Parameters
    ----------
    notebook : bool
        Whether the comment is for a notebook.
    rules : Optional[List[str]], optional
        The pylint rules to ignore, by default None.

    Returns
    -------
    str
        The pylint ignore comment string.

    Example
    -------
    ```python
    >>> get_pylint_ignore_comment(True, ["invalid-name", "line-too-long"])

    # pylint: disable=invalid-name, line-too-long
    ```
    """
    if not rules:
        rules = PYLINT_RULES
    line = "# pylint: disable=" + ",".join(rules)
    if notebook is True:
        line = "\n" + line
    return line + "\n"



"""Waldiez model."""

from .model import WaldiezModel
from .model_data import WaldiezModelAPIType, WaldiezModelData, WaldiezModelPrice

__all__ = [
    "WaldiezModel",
    "WaldiezModelData",
    "WaldiezModelPrice",
    "WaldiezModelAPIType",
]



"""Waldiez chat model."""

from typing import Any, Dict, Optional

from pydantic import Field
from typing_extensions import Annotated

from ..agents import WaldiezAgent, WaldiezRagUser
from ..common import WaldiezBase
from .chat_data import WaldiezChatData
from .chat_message import WaldiezChatMessage
from .chat_nested import WaldiezChatNested


class WaldiezChat(WaldiezBase):
    """Chat class.

    Attributes
    ----------
    id : str
        The chat ID.
    data : WaldiezChatData
        The chat data.
        See `waldiez.models.chat.WaldiezChatData` for more information.
    name : str
        The chat name.
    source : str
        The chat source.
    target : str
        The chat target.
    nested_chat : WaldiezChatNested
        The nested chat message/reply if any.
    message : WaldiezChatMessage
        The chat message.
    message_content : Optional[str]
        The chat message content if any. If method, the method's body.

    Functions
    ---------
    get_chat_args()
        Get the chat arguments to use in autogen.
    """

    id: Annotated[
        str,
        Field(
            ...,
            title="ID",
            description="The chat ID.",
        ),
    ]
    data: Annotated[
        WaldiezChatData,
        Field(
            ...,
            title="Data",
            description="The chat data.",
        ),
    ]

    @property
    def name(self) -> str:
        """Get the name."""
        return self.data.name

    @property
    def source(self) -> str:
        """Get the source."""
        if self.data.real_source:
            return self.data.real_source
        return self.data.source

    @property
    def target(self) -> str:
        """Get the target."""
        if self.data.real_target:
            return self.data.real_target
        return self.data.target

    @property
    def nested_chat(self) -> WaldiezChatNested:
        """Get the nested chat."""
        return self.data.nested_chat

    @property
    def message(self) -> WaldiezChatMessage:
        """Get the message."""
        if isinstance(
            self.data.message, str
        ):  # pragma: no cover (just for the lint)
            return WaldiezChatMessage(
                type="string",
                use_carryover=False,
                content=self.data.message,
                context={},
            )
        return self.data.message

    @property
    def message_content(self) -> Optional[str]:
        """Get the message content."""
        return self.data.message_content

    def get_chat_args(
        self,
        sender: Optional[WaldiezAgent] = None,
    ) -> Dict[str, Any]:
        """Get the chat arguments to use in autogen.

        Parameters
        ----------
        sender : WaldiezAgent
            The sender agent, to check if it's a RAG user.
        Returns
        -------
        dict
            The chat arguments.
        """
        args_dict = self.data.get_chat_args()
        if (
            isinstance(sender, WaldiezRagUser)
            and sender.agent_type == "rag_user"
            and self.message.type == "rag_message_generator"
        ):
            # check for n_results in agent data, to add in context
            n_results = sender.data.retrieve_config.n_results
            if isinstance(n_results, int) and n_results > 0:
                args_dict["n_results"] = n_results
        return args_dict



"""Agent skills related string generation functions."""

from typing import Dict, List

from waldiez.models import WaldiezAgent, WaldiezSkill

from ..utils import get_escaped_string


def get_agent_skill_registrations(
    agent: WaldiezAgent,
    agent_names: Dict[str, str],
    all_skills: List[WaldiezSkill],
    skill_names: Dict[str, str],
) -> str:
    """Get the agent skill registrations.

    example output:

    ```python
    >>> register_function(
        {skill_name},
        caller={agent_name},
        executor={executor_agent_name},
        name="{skill_name}",
        description="{skill_description}",
    )
    ```

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    agent_names : Dict[str, str]
        A mapping of agent id to agent name.
    all_skills : List[WaldiezSkill]
        All the skills in the flow.
    skill_names : Dict[str, str]
        A mapping of skill id to skill name.

    Returns
    -------
    str
        The agent skill registrations.
    """
    if not agent.data.skills or not all_skills:
        return ""
    content = ""
    for linked_skill in agent.data.skills:
        skill_name = skill_names[linked_skill.id]
        waldiez_skill = next(
            skill for skill in all_skills if skill.id == linked_skill.id
        )
        skill_description = (
            waldiez_skill.description or f"Description of {skill_name}"
        )
        skill_description = get_escaped_string(skill_description)
        content += (
            f"register_function(\n"
            f"    {skill_name},\n"
            f"    caller={agent_names[agent.id]},\n"
            f"    executor={agent_names[linked_skill.executor_id]},\n"
            f'    name="{skill_name}",\n'
            f'    description="{skill_description}",\n'
            f")\n\n"
        )
    return content



"""Function related utilities."""

import ast
from typing import Dict, List, Literal, Optional, Tuple

WaldiezMethodName = Literal[
    "callable_message",  # Chat
    "is_termination_message",  # Agent
    "nested_chat_message",  # Agents' NestedChat
    "nested_chat_reply",  # Agents' NestedChat
    "custom_speaker_selection",  # GroupChat
    "custom_embedding_function",  # RAG
    "custom_token_count_function",  # RAG
    "custom_text_split_function",  # RAG
]

METHOD_ARGS: Dict[WaldiezMethodName, List[str]] = {
    "callable_message": ["sender", "recipient", "context"],
    "is_termination_message": ["message"],
    "nested_chat_message": ["recipient", "messages", "sender", "config"],
    "nested_chat_reply": ["recipient", "messages", "sender", "config"],
    "custom_speaker_selection": ["last_speaker", "groupchat"],
    "custom_embedding_function": [],
    "custom_token_count_function": ["text", "model"],
    "custom_text_split_function": [
        "text",
        "max_tokens",
        "chunk_mode",
        "must_break_at_empty_line",
        "overlap",
    ],
}

# pylint: disable=line-too-long
METHOD_TYPE_HINTS: Dict[WaldiezMethodName, str] = {
    "callable_message": "# type: (ConversableAgent, ConversableAgent, dict) -> Union[dict, str]",
    "is_termination_message": "# type: (dict) -> bool",
    "nested_chat_message": "# type: (ConversableAgent, list[dict], ConversableAgent, dict) -> Union[dict, str]",
    "nested_chat_reply": "# type: (ConversableAgent, list[dict], ConversableAgent, dict) -> Union[dict, str]",
    "custom_speaker_selection": "# type: (ConversableAgent, GroupChat) -> Union[Agent, str, None]",
    "custom_embedding_function": "# type: () -> Callable[..., Any]",
    "custom_token_count_function": "# type: (str, str) -> int",
    "custom_text_split_function": "# type: (str, int, str, bool, int) -> List[str]",
}


def parse_code_string(
    code_string: str,
) -> Tuple[Optional[str], Optional[ast.Module]]:
    """Parse the code string.

    Parameters
    ----------
    code_string : str
        The code string.

    Returns
    -------
    Tuple[Optional[str], Optional[ast.Module]]
        If valid, None and the ast module.
        If invalid, the error message and None.
    """
    # pylint: disable=broad-except
    try:
        tree = ast.parse(code_string)
    except SyntaxError as e:
        return f"SyntaxError: {e}, in \n{code_string}", None
    except BaseException as e:  # pragma: no cover
        return f"Invalid code: {e}, in \n{code_string}", None
    return None, tree


def check_function(
    code_string: str,
    function_name: WaldiezMethodName,
    skip_type_hints: bool = False,
) -> Tuple[bool, str]:
    """Check the function.

    Parameters
    ----------
    code_string : str
        The code string.
    function_name : WaldiezMethodName
        The expected function name.
    skip_type_hints : bool, optional
        Whether to skip type hints in the function body, by default False.

    Returns
    -------
    Tuple[bool, str]
        If valid, True and the function body (only), no extra lines.
        If invalid, False and the error message.
    """
    error, tree = parse_code_string(code_string)
    if error is not None or tree is None:
        return False, error or "Invalid code"
    if function_name not in METHOD_ARGS:
        return False, f"Invalid function name: {function_name}"
    expected_method_args = METHOD_ARGS[function_name]
    return _get_function_body(
        tree,
        code_string,
        function_name,
        expected_method_args,
        skip_type_hints=skip_type_hints,
    )


def _get_function_body(
    tree: ast.Module,
    code_string: str,
    function_name: WaldiezMethodName,
    method_args: List[str],
    skip_type_hints: bool = False,
) -> Tuple[bool, str]:
    """Get the function body.

    Parameters
    ----------
    tree : ast.Module
        The ast module.
    code_string : str
        The code string.
    function_name : WaldiezMethodName
        The expected function name.
    method_args : List[str]
        The expected method arguments.

    Returns
    -------
    Tuple[bool, str]
        If valid, True and the function body (only), no extra lines.
        If invalid, False and the error message.
    """
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            if node.name != function_name:
                continue
            if len(node.args.args) != len(method_args):
                return (
                    False,
                    f"Invalid number of arguments in function {node.name}",
                )
            for arg, expected_arg in zip(node.args.args, method_args):
                if arg.arg != expected_arg:
                    return (
                        False,
                        f"Invalid argument name in function {node.name}",
                    )
            function_body_lines = code_string.splitlines()[
                node.lineno - 1 : node.end_lineno
            ]
            function_body = "\n".join(function_body_lines[1:])
            if not skip_type_hints:
                # add type hints after the function definition
                function_body = (
                    f"    {METHOD_TYPE_HINTS[function_name]}\n{function_body}"
                )
            return True, function_body
    error_msg = (
        f"No function with name `{function_name}`"
        f" and arguments `{method_args}` found"
    )
    return False, error_msg



"""RAG user agent.
# pylint: disable=line-too-long
It extends a user agent and has RAG related parameters.
"""

from .rag_user import WaldiezRagUser
from .rag_user_data import WaldiezRagUserData
from .retrieve_config import (
    WaldiezRagUserChunkMode,
    WaldiezRagUserModels,
    WaldiezRagUserRetrieveConfig,
    WaldiezRagUserTask,
    WaldiezRagUserVectorDb,
)
from .vector_db_config import WaldiezRagUserVectorDbConfig

__all__ = [
    "WaldiezRagUser",
    "WaldiezRagUserData",
    "WaldiezRagUserModels",
    "WaldiezRagUserVectorDb",
    "WaldiezRagUserChunkMode",
    "WaldiezRagUserRetrieveConfig",
    "WaldiezRagUserTask",
    "WaldiezRagUserVectorDbConfig",
]



"""Waldiez Message Model."""

from typing import Any, Dict, Optional, Tuple, Union

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..common import WaldiezBase, WaldiezMethodName, check_function

WaldiezChatMessageType = Literal[
    "string", "method", "rag_message_generator", "none"
]


class WaldiezChatMessage(WaldiezBase):
    """
    Waldiez Message.

    A generic message with a type and content.

    If the type is not `none`, the content is a string.
    If the type is 'method', the content is the source code of a method.
    If the type is 'last_carryover', the content is a method to return
        the last carryover from the context.
    If the type is 'rag_message_generator', and the sender is a RAG user agent,
        the content will be generated by the `sender.message_generator` method.

    Attributes
    ----------
    type : WaldiezChatMessageType
        The type of the message:
        - string
        - method
        - rag_message_generator
        - none
        If the sender is a RAG user agent,
        and the type is `rag_message_generator`,
        the `{sender}.message_generator` method will be used.
    content : Optional[str]
        The content of the message (string or method).
    context : Dict[str, Any]
        Extra context of the message.
    """

    type: Annotated[
        WaldiezChatMessageType,
        Field(
            "none",
            title="Type",
            description=(
                "The type of the message: "
                "`string`, `method`, "
                "`rag_message_generator` or `none`."
                "If last_carryover, a method to return the context's"
                "last carryover will be used."
                "If the sender is a RAG user agent, "
                "and the type is `message_generator`,"
                "the `sender.message_generator` method will be used."
            ),
        ),
    ]
    use_carryover: Annotated[
        bool,
        Field(
            False,
            title="Use Carryover",
            description="Use the carryover from the context.",
        ),
    ]
    content: Annotated[
        Optional[str],
        Field(
            None,
            title="Content",
            description="The content of the message (string or method).",
        ),
    ]
    context: Annotated[
        Dict[str, Any],
        Field(
            default_factory=dict,
            title="Context",
            description="Extra context of the message.",
        ),
    ]


def validate_message_dict(
    value: Dict[
        Literal["type", "use_carryover", "content", "context"],
        Union[Optional[str], Optional[bool], Optional[Dict[str, Any]]],
    ],
    function_name: WaldiezMethodName,
    skip_definition: bool = False,
) -> WaldiezChatMessage:
    """Validate a message dict.

    Check the provided message dict.
    Depending on the type, the content is validated.
    If the type is "method", the content is checked against the function name.

    Parameters
    ----------
    value : dict
        The message dict.
    function_name : str (WaldiezMethodName)
        The function name.
    skip_definition : bool, optional
        Skip the function definition in the content, by default False

    Returns
    -------
    WaldiezChatMessage
        The validated message.

    Raises
    ------
    ValueError
        If the validation fails.
    """
    message_type, use_carryover, content, context = _get_message_args_from_dict(
        value
    )
    if message_type == "string":
        if not content or not isinstance(content, str):
            content = ""
        if use_carryover:
            method_content = _get_last_carryover_method_content(content)
            return WaldiezChatMessage(
                type="method",
                use_carryover=True,
                content=method_content,
                context=context,
            )
        return WaldiezChatMessage(
            type="string",
            use_carryover=False,
            content=content,
            context=context,
        )
    if message_type == "none":
        return WaldiezChatMessage(
            type="none",
            use_carryover=False,
            content=None,
            context=context,
        )
    if message_type == "method":
        if not content:
            raise ValueError(
                "The message content is required for the method type"
            )
        valid, error_or_content = check_function(
            content, function_name, skip_type_hints=use_carryover
        )
        if not valid:
            raise ValueError(error_or_content)
        message_content = error_or_content if skip_definition else content
        return WaldiezChatMessage(
            type="method",
            use_carryover=use_carryover,
            content=message_content,
            context=context,
        )
    if message_type == "rag_message_generator":
        if use_carryover:
            return WaldiezChatMessage(
                type="method",
                use_carryover=True,
                content=RAG_METHOD_WITH_CARRYOVER,
                context=context,
            )
        return WaldiezChatMessage(
            type="rag_message_generator",
            use_carryover=use_carryover,
            content=None,
            context=context,
        )
    raise ValueError("Invalid message type")  # pragma: no cover


def _get_message_args_from_dict(
    value: Dict[
        Literal["type", "use_carryover", "content", "context"],
        Union[Optional[str], Optional[bool], Optional[Dict[str, Any]]],
    ],
) -> Tuple[str, bool, Optional[str], Dict[str, Any]]:
    """Get the message args from a dict.

    Parameters
    ----------
    value : dict
        The message dict.

    Returns
    -------
    tuple
        The message type, content, and context.

    Raises
    ------
    ValueError
        If the message type is invalid.
    """
    message_type = value.get("type")
    if not isinstance(message_type, str) or message_type not in (
        "string",
        "method",
        "rag_message_generator",
        "none",
    ):
        raise ValueError("Invalid message type")
    use_carryover = value.get("use_carryover", False)
    if not isinstance(use_carryover, bool):
        use_carryover = False
    content = value.get("content", "")
    if not isinstance(content, str):
        content = ""
    context: Dict[str, Any] = {}
    context_value = value.get("context")
    if isinstance(context_value, dict):
        context = context_value
    if not isinstance(context, dict):  # pragma: no cover
        context = {}
    return message_type, use_carryover, content, context


def _get_last_carryover_method_content(text_content: str) -> str:
    """Get the last carryover method content.

    Parameters
    ----------
    text_content : str
        Text content before the carryover.
    Returns
    -------
    str
        The last carryover method content.
    """
    method_content = '''
def callable_message(sender, recipient, context):
    # type: (ConversableAgent, ConversableAgent, dict) -> Union[dict, str]
    """Get the message to send using the last carryover.

    Parameters
    ----------
    sender : ConversableAgent
        The source agent.
    recipient : ConversableAgent
        The target agent.
    context : dict
        The context.

    Returns
    -------
    Union[dict, str]
        The message to send using the last carryover.
    """
    carryover = context.get("carryover", "")
    if isinstance(carryover, list):
        carryover = carryover[-1]
    if not isinstance(carryover, str):
        carryover = ""'''
    if text_content:
        method_content += f"""
    final_message = "{text_content}" + carryover
    return final_message
"""
    else:
        method_content += """
    return carryover
"""
    return method_content


RAG_METHOD_WITH_CARRYOVER = '''
def callable_message(sender, recipient, context):
    # type: (RetrieveUserProxyAgent, ConversableAgent, dict) -> Union[dict, str]
    """Get the message using the RAG message generator method.

    Parameters
    ----------
    sender : RetrieveUserProxyAgent
        The source agent.
    recipient : ConversableAgent
        The target agent.
    context : dict
        The context.

    Returns
    -------
    Union[dict, str]
        The message to send using the last carryover.
    """
    carryover = context.get("carryover", "")
    if isinstance(carryover, list):
        carryover = carryover[-1]
    if not isinstance(carryover, str):
        carryover = ""
    message = sender.message_generator(sender, recipient, context)
    if carryover:
        message += carryover
    return message
'''



"""Tools for exporting agents, models, skills and chats to strings."""

from .flow import export_flow
from .models import export_models
from .skills import export_skills
from .utils import comment, get_valid_instance_name

__all__ = [
    "export_flow",
    "comment",
    "get_valid_instance_name",
    "export_models",
    "export_skills",
]



"""Method related string generation utilities."""

from waldiez.models import METHOD_ARGS, WaldiezMethodName


def get_method_string(
    method_name: WaldiezMethodName, renamed_method_name: str, method_body: str
) -> str:
    """Get a function string.

    Parameters
    ----------
    method_name : WaldiezMethodName
        The method name.
    renamed_method_name : str
        The renamed method name.
    method_body : str
        The method body.

    Returns
    -------
    str
        The function string having the definition, type hints and body.
    """
    method_args = METHOD_ARGS[method_name]
    content = f"def {renamed_method_name}("
    if len(method_args) == 0:
        content += "):"
    else:
        content += "\n"
        for arg in method_args:
            content += f"    {arg},\n"
        content += "):"
    content += f"\n{method_body}"
    return content



"""Waldiez flow data."""

from typing import Any, Dict, List

from pydantic import Field
from typing_extensions import Annotated

from ..agents import WaldiezAgents
from ..chat import WaldiezChat
from ..common import WaldiezBase
from ..model import WaldiezModel
from ..skill import WaldiezSkill


class WaldiezFlowData(WaldiezBase):
    """Flow data class.

    Attributes
    ----------
    nodes : List[Dict[str, Any]]
        The nodes of the flow. We ignore this (UI-related)
    edges : List[Dict[str, Any]]
        The edges of the flow. We ignore this (UI-related)
    viewport : Dict[str, Any]
        The viewport of the flow. We ignore this (UI-related)
    agents : WaldiezAgents
        The agents of the flow:
        users: List[WaldiezUserProxy]
        assistants: List[WaldiezAssistant]
        managers: List[WaldiezGroupManager]
        rag_users : List[WaldiezRagUser]
        See `WaldiezAgents` for more info.
    models : List[WaldiezModel]
        The models of the flow. See `WaldiezModel`.
    skills : List[WaldiezSkill]
        The skills of the flow. See `WaldiezSkill`.
    chats : List[WaldiezChat]
        The chats of the flow. See `WaldiezChat`.
    """

    # the ones below (nodes,edges, viewport) we ignore
    # (they for graph connections, positions, etc.)
    nodes: Annotated[
        List[Dict[str, Any]],
        Field(default_factory=list),
    ]
    edges: Annotated[
        List[Dict[str, Any]],
        Field(default_factory=list),
    ]
    viewport: Annotated[
        Dict[str, Any],
        Field(default_factory=dict),
    ]
    # these are the ones we use.
    agents: Annotated[
        WaldiezAgents,
        Field(
            description="The agents of the flow",
            title="Agents",
            default_factory=WaldiezAgents,
        ),
    ]
    models: Annotated[
        List[WaldiezModel],
        Field(
            description="The models of the flow",
            title="Models",
            default_factory=list,
        ),
    ]
    skills: Annotated[
        List[WaldiezSkill],
        Field(
            description="The skills of the flow",
            title="Skills",
            default_factory=list,
        ),
    ]
    chats: Annotated[
        List[WaldiezChat],
        Field(
            description="The chats of the flow",
            title="Chats",
            default_factory=list,
        ),
    ]



"""Custom IOStream class to use with autogen.

It is meant to be used when we want to use custom
`print` and `input`. For example, when a websocket
is used to trigger a UI element that requires user input.
and sends back the user's input to the websocket. In the same
way, we can use it to forward what is meant to be printed.
"""

import threading
from typing import Any, Callable, Optional

from autogen.io import IOStream  # type: ignore[import-untyped]


class WaldiezIOStream(IOStream):
    """Custom IOStream class to handle `print` and `input` functions."""

    def __init__(
        self,
        input_timeout: float = 60.0,
        print_function: Optional[Callable[..., None]] = None,
        on_prompt_input: Optional[Callable[[str], None]] = None,
    ) -> None:
        """
        Initialize the IOStream.

        Parameters
        ----------
        input_timeout : float, optional
            The input timeout in seconds, by default 60.0.
        print_function : Optional[Callable[..., None]], optional
            The function to handle print operations, by default None.
        on_prompt_input : Optional[Callable[[str], None]], optional
            The function to call for processing input prompts, by default None.

        Notes
        -----
        - on_prompt_input: It does not return a string (like 'input' does).
            Instead, it is meant to be used to forward the prompt somewhere else
            (e.g., a websocket). When we get the input, we can call
            `waldiez_io_stream.set_input(input_data)` with the input data.
        """
        self.input_timeout = input_timeout  # Timeout for input
        self.print_function = print_function  # Custom print handler
        self._on_prompt_input = on_prompt_input  # Custom input prompt handler
        self.current_input: Optional[str] = None  # Store the current input
        self._input_event = threading.Event()  # Event to signal input readiness
        self.allow_input = True  # Flag to allow or block input setting

    def print(
        self,
        *objects: Any,
        sep: str = " ",
        end: str = "\n",
        flush: bool = False,
    ) -> None:
        """
        Mock the `print` function.

        Parameters
        ----------
        objects : Any
            The objects to print.
        sep : str, optional
            The separator, by default " ".
        end : str, optional
            The ending string, by default a new line.
        flush : bool, optional
            Whether to flush the output, by default False.
        """
        print_function: Callable[..., None] = self.print_function or print
        print_function(*objects, sep=sep, end=end, flush=flush)

    def input(self, prompt: str = "", *, password: bool = False) -> str:
        """
        Mock the `input` function with optional timeout handling.

        Parameters
        ----------
        prompt : str, optional
            The prompt to show, by default "".
        password : bool, optional
            Whether to hide the input as password (not used), by default False.

        Returns
        -------
        str
            The user's input or '\n' if timeout occurs.
        """
        _prompt = prompt or "Your input:"
        if _prompt in (">", "> "):  # pragma: no cover
            _prompt = "Your input:"
        if prompt:
            if self._on_prompt_input:
                self._on_prompt_input(_prompt)
            self.print(_prompt, end="")

        # Only reset if no input is currently set
        # e.g. handle the case when we call set_input(..) before input(..)
        already_set = self._input_event.is_set()
        if not self._input_event.is_set():
            self.current_input = None  # Reset previous input
            self._input_event.clear()  # Clear the event before waiting

        # Wait for input or timeout
        if not self._input_event.wait(self.input_timeout):
            # Timeout occurred, return what we have so far
            to_return = (
                self.current_input if self.current_input is not None else "\n"
            )
            self.current_input = None
            self._input_event.clear()
            # if we had already set the input, return it
            return to_return if already_set else "\n"

        # Input is ready, return it
        to_return = (
            self.current_input if self.current_input is not None else "\n"
        )
        self.current_input = None
        self._input_event.clear()  # Clear the event after waiting
        return to_return

    def set_input(self, value: str) -> None:
        """
        Set the input value and signal that input is ready.

        Parameters
        ----------
        value : str
            The value to set as input.
        """
        if self.allow_input:  # Respect the allow_input flag
            self.current_input = value
            self._input_event.set()  # Signal that input is ready

    def reset_state(self) -> None:
        """Reset the IOStream state for testing."""
        self.current_input = None
        self._input_event.clear()
        self.allow_input = True  # Re-enable input setting



"""Waldiez package."""

from ._version import __version__
from .conflict_checker import check_conflicts
from .exporter import WaldiezExporter
from .models import Waldiez
from .runner import WaldiezRunner

# flag to check if ag2 and autogen-agentchat
# are installed at the same time
__WALDIEZ_CHECKED_FOR_CONFLICTS = False


def _check_conflicts_once() -> None:
    """Check for conflicts once."""
    # pylint: disable=global-statement
    global __WALDIEZ_CHECKED_FOR_CONFLICTS
    if __WALDIEZ_CHECKED_FOR_CONFLICTS is False:
        check_conflicts()
        __WALDIEZ_CHECKED_FOR_CONFLICTS = True


_check_conflicts_once()


__all__ = [
    "Waldiez",
    "WaldiezExporter",
    "WaldiezRunner",
    "__version__",
]



"""Waldiez exporter.

The role of the exporter is to export the model's data
to an autogen's flow with one or more chats.

The resulting file(s): a `flow.py` file with one `main()` function
to trigger the chat(s).
If additional tools/skills are used,
they are exported as their `skill_name` in the same directory with
the `flow.py` file. So the `flow.py` could have entries like:
`form {skill1_name} import {skill1_name}`
`form {skill2_name} import {skill2_name}`
"""

# pylint: disable=inconsistent-quotes

import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Dict, List, Optional, Union

from .exporting import comment, export_flow, get_valid_instance_name
from .models import (
    Waldiez,
    WaldiezAgent,
    WaldiezChat,
    WaldiezModel,
    WaldiezSkill,
)


class WaldiezExporter:
    """Waldiez exporter.

    Attributes:
        waldiez (Waldiez): The Waldiez instance.
    """

    _agent_names: Dict[str, str]
    _model_names: Dict[str, str]
    _skill_names: Dict[str, str]
    _chat_names: Dict[str, str]
    _chats: List[WaldiezChat]
    _skills: List[WaldiezSkill]
    _models: List[WaldiezModel]
    _agents: List[WaldiezAgent]

    def __init__(self, waldiez: Waldiez) -> None:
        """Initialize the Waldiez exporter.

        Parameters:
            waldiez (Waldiez): The Waldiez instance.
        """
        self.waldiez = waldiez
        self._initialize()

    @classmethod
    def load(cls, file_path: Path) -> "WaldiezExporter":
        """Load the Waldiez instance from a file.

        Parameters
        ----------
        file_path : Path
            The file path.

        Returns
        -------
        WaldiezExporter
            The Waldiez exporter.
        """
        waldiez = Waldiez.load(file_path)
        return cls(waldiez)

    def _initialize(
        self,
    ) -> None:
        """Get all the names in the flow.

        We need to make sure that no duplicate names are used,
        and that the names can be used as python variables.
        """
        all_names: Dict[str, str] = {}
        agent_names: Dict[str, str] = {}
        model_names: Dict[str, str] = {}
        skill_names: Dict[str, str] = {}
        chat_names: Dict[str, str] = {}
        chats: List[WaldiezChat] = []
        skills: List[WaldiezSkill] = []
        models: List[WaldiezModel] = []
        agents: List[WaldiezAgent] = []
        for agent in self.waldiez.agents:
            all_names = get_valid_instance_name(
                (agent.id, agent.name), all_names, prefix="wa"
            )
            agent_names[agent.id] = all_names[agent.id]
            agents.append(agent)
        for model in self.waldiez.models:
            all_names = get_valid_instance_name(
                (model.id, model.name), all_names, prefix="wm"
            )
            model_names[model.id] = all_names[model.id]
            models.append(model)
        for skill in self.waldiez.skills:
            all_names = get_valid_instance_name(
                (skill.id, skill.name), all_names, prefix="ws"
            )
            skill_names[skill.id] = all_names[skill.id]
            skills.append(skill)
        for chat in self.waldiez.flow.data.chats:
            all_names = get_valid_instance_name(
                (chat.id, chat.name), all_names, prefix="wc"
            )
            chat_names[chat.id] = all_names[chat.id]
            chats.append(chat)
        self._agent_names = agent_names
        self._model_names = model_names
        self._skill_names = skill_names
        self._chat_names = chat_names
        self._chats = chats
        self._skills = skills
        self._models = models
        self._agents = agents

    def export(self, path: Union[str, Path], force: bool = False) -> None:
        """Export the Waldiez instance.

        Parameters
        ----------
        path : Union[str, Path]
            The path to export to.
        force : bool, optional
            Override the output file if it already exists, by default False.

        Raises
        ------
        FileExistsError
            If the file already exists and force is False.
        IsADirectoryError
            If the output is a directory.
        ValueError
            If the file extension is invalid.
        """
        if not isinstance(path, Path):
            path = Path(path)
        path = path.resolve()
        if path.is_dir():
            raise IsADirectoryError(f"Output is a directory: {path}")
        if path.exists():
            if force is False:
                raise FileExistsError(f"File already exists: {path}")
            path.unlink(missing_ok=True)
        path.parent.mkdir(parents=True, exist_ok=True)
        extension = path.suffix
        if extension == ".waldiez":
            self.to_waldiez(path)
        elif extension == ".py":
            self.to_py(path)
        elif extension == ".ipynb":
            self.to_ipynb(path)
        else:
            raise ValueError(f"Invalid extension: {extension}")

    def to_ipynb(self, path: Path) -> None:
        """Export flow to jupyter notebook.

        Parameters
        ----------
        path : Path
            The path to export to.

        Raises
        ------
        RuntimeError
            If the notebook could not be generated.
        """
        content = f"{comment(True)}{self.waldiez.name}" + "\n\n"
        content += f"{comment(True, 2)}Dependencies" + "\n\n"
        content += "import sys\n"
        requirements = " ".join(self.waldiez.requirements)
        if requirements:
            content += (
                f"# !{{sys.executable}} -m pip install -q {requirements}" + "\n"
            )
        content += export_flow(
            waldiez=self.waldiez,
            agents=(self._agents, self._agent_names),
            chats=(self._chats, self._chat_names),
            models=(self._models, self._model_names),
            skills=(self._skills, self._skill_names),
            output_dir=path.parent,
            notebook=True,
        )
        # we first create a .py file with the content
        # and then convert it to a notebook using jupytext
        py_path = path.with_suffix(".tmp.py")
        with open(py_path, "w", encoding="utf-8") as f:
            f.write(content)
        if not shutil.which("jupytext"):  # pragma: no cover
            run_command(
                [sys.executable, "-m", "pip", "install", "jupytext"],
                allow_error=False,
            )
        run_command(
            ["jupytext", "--to", "notebook", str(py_path)],
            allow_error=False,
        )
        ipynb_path = str(py_path).replace(".tmp.py", ".tmp.ipynb")
        if not os.path.exists(ipynb_path):  # pragma: no cover
            raise RuntimeError("Could not generate notebook")
        Path(ipynb_path).rename(ipynb_path.replace(".tmp.ipynb", ".ipynb"))
        py_path.unlink(missing_ok=True)

    def to_py(self, path: Path) -> None:
        """Export waldiez flow to python script.

        Parameters
        ----------
        path : Path
            The path to export to.
        """
        content = "#!/usr/bin/env python\n"
        content += f'"""{self.waldiez.name}\n\n'
        content += f"{self.waldiez.description}\n\n"
        content += f"Tags: {', '.join(self.waldiez.tags)}\n\n"
        content += f"Requirements: {', '.join(self.waldiez.requirements)}\n\n"
        content += '"""\n\n'
        content += "# cspell: disable\n"
        content += "# flake8: noqa\n\n"
        content += export_flow(
            waldiez=self.waldiez,
            agents=(self._agents, self._agent_names),
            chats=(self._chats, self._chat_names),
            models=(self._models, self._model_names),
            skills=(self._skills, self._skill_names),
            output_dir=path.parent,
            notebook=False,
        )
        content += '\n\nif __name__ == "__main__":\n'
        content += "    print(main())\n"
        with open(path, "w", encoding="utf-8") as file:
            file.write(content)

    def to_waldiez(self, file_path: Path) -> None:
        """Export the Waldiez instance.

        Parameters
        ----------
        file_path : Path
            The file path.
        """
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(self.waldiez.model_dump_json())


def run_command(
    cmd: List[str],
    cwd: Optional[Path] = None,
    allow_error: bool = True,
) -> None:
    """Run a command.

    Parameters
    ----------
    cmd : List[str]
        The command to run.
    cwd : Path, optional
        The working directory, by default None (current working directory).
    allow_error : bool, optional
        Whether to allow errors, by default True.

    Raises
    ------
    RuntimeError
        If the command fails and allow_error is False.
    """
    if not cwd:
        cwd = Path.cwd()
    # pylint: disable=broad-except
    try:
        subprocess.run(
            cmd,
            check=True,
            cwd=cwd,
            env=os.environ,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )  # nosemgrep # nosec
    except BaseException as error:  # pragma: no cover
        if allow_error:
            return
        raise RuntimeError(f"Error running command: {error}") from error
